@@ -2,13 +2,18 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: array.h,v 1.44 2002/10/11 01:39:28 nilsson Exp $
+|| $Id: array.h,v 1.63 2004/09/28 23:58:35 mast Exp $
 */
 
 #ifndef ARRAY_H
 #define ARRAY_H
 
-#include "las.h"
+#include "svalue.h"
+#include "dmalloc.h"
+
+/* This debug tool writes out messages whenever arrays with unfinished
+ * type fields are encountered. */
+/* #define TRACE_UNFINISHED_TYPE_FIELDS */
 
 struct array
 {
@@ -19,10 +24,11 @@ struct array
 			 * when freeing arrays */
   INT32 size;		/* number of items in this array */
   INT32 malloced_size;	/* number of elements that can fit in this array */
-  TYPE_FIELD type_field;/* A bitfield with one bit for each type.
-			 * Bits can be set that don't exist in the array
-			 * though.
-			 */
+  TYPE_FIELD type_field;/* A bitfield with one bit for each type. */
+  /* Bits can be set that don't exist in the array. type_field is
+   * initialized to BIT_MIXED|BIT_UNFINISHED for newly allocated
+   * arrays so that they can be modified without having to update
+   * this. It should be set accurately when that's done, though. */
   INT16 flags;          /* ARRAY_* flags */
   struct svalue *item;
   struct svalue real_item[1];
@@ -34,6 +40,7 @@ struct array
 #define ARRAY_WEAK_SHRINK 8
 
 PMOD_EXPORT extern struct array empty_array, weak_empty_array, weak_shrink_empty_array;
+extern struct array *first_array;
 extern struct array *gc_internal_array;
 
 #if defined(DEBUG_MALLOC) && defined(PIKE_DEBUG)
@@ -42,19 +49,6 @@ extern struct array *gc_internal_array;
 #define ITEM(X) ((X)->item)
 #endif
 
-#define LINK_ARRAY(a) do {						\
-  (a)->prev = &empty_array;						\
-  (a)->next = empty_array.next;						\
-  empty_array.next = (a);						\
-  (a)->next->prev = (a);						\
-} while (0)
-
-#define UNLINK_ARRAY(a) do {						\
-  struct array *next = (a)->next, *prev = (a)->prev;			\
-  prev->next = next;							\
-  next->prev = prev;							\
-} while (0)
-
 /* These are arguments for the function 'merge' which merges two sorted
  * set stored in arrays in the way you specify
  */
@@ -75,7 +69,16 @@ extern struct array *gc_internal_array;
 #define PIKE_ARRAY_OP_SUB PIKE_MINTERM(PIKE_ARRAY_OP_TAKE_A,PIKE_ARRAY_OP_SKIP_A ,PIKE_ARRAY_OP_SKIP_B)
 
 
-#define free_array(V) do{ struct array *v_=(V); debug_malloc_touch(v_); if(!sub_ref(v_)) really_free_array(v_); }while(0)
+#define free_array(V) do{						\
+    struct array *v_=(V);						\
+    debug_malloc_touch(v_);						\
+    DO_IF_DEBUG (							\
+      DO_IF_PIKE_CLEANUP (						\
+	if (gc_external_refs_zapped)					\
+	  gc_check_zapped (v_, PIKE_T_ARRAY, __FILE__, __LINE__)));	\
+    if(!sub_ref(v_))							\
+      really_free_array(v_);						\
+  }while(0)
 
 #define allocate_array(X) low_allocate_array((X),0)
 #define allocate_array_no_init(X,Y) low_allocate_array((X),(Y))
@@ -91,6 +94,8 @@ PMOD_EXPORT void really_free_array(struct array *v);
 PMOD_EXPORT void do_free_array(struct array *a);
 PMOD_EXPORT struct array *array_set_flags(struct array *a, int flags);
 PMOD_EXPORT void array_index(struct svalue *s,struct array *v,INT32 ind);
+PMOD_EXPORT struct array *array_column (struct array *data, struct svalue *index,
+					int destructive);
 PMOD_EXPORT void simple_array_index_no_free(struct svalue *s,
 				struct array *a,struct svalue *ind);
 PMOD_EXPORT void array_free_index(struct array *v,INT32 ind);
@@ -112,6 +117,7 @@ PMOD_EXPORT INT32 array_find_destructed_object(struct array *v);
 INT32 *get_order(struct array *v, cmpfun fun);
 INLINE int set_svalue_cmpfun(const struct svalue *a, const struct svalue *b);
 PMOD_EXPORT void sort_array_destructively(struct array *v);
+PMOD_EXPORT INT32 *stable_sort_array_destructively(struct array *v);
 PMOD_EXPORT INT32 *get_set_order(struct array *a);
 PMOD_EXPORT INT32 *get_switch_order(struct array *a);
 PMOD_EXPORT INT32 *get_alpha_order(struct array *a);
@@ -119,9 +125,10 @@ INT32 set_lookup(struct array *a, struct svalue *s);
 INT32 switch_lookup(struct array *a, struct svalue *s);
 PMOD_EXPORT struct array *order_array(struct array *v, INT32 *order);
 PMOD_EXPORT struct array *reorder_and_copy_array(struct array *v, INT32 *order);
-PMOD_EXPORT void array_fix_type_field(struct array *v);
+PMOD_EXPORT TYPE_FIELD array_fix_type_field(struct array *v);
+#ifdef PIKE_DEBUG
 void array_check_type_field(struct array *v);
-PMOD_EXPORT struct array *compact_array(struct array *v);
+#endif
 PMOD_EXPORT union anything *low_array_get_item_ptr(struct array *a,
 				       INT32 ind,
 				       TYPE_T t);
@@ -145,11 +152,13 @@ int array_is_constant(struct array *a,
 node *make_node_from_array(struct array *a);
 PMOD_EXPORT void push_array_items(struct array *a);
 void describe_array_low(struct array *a, struct processing *p, int indent);
-PMOD_EXPORT void simple_describe_array(struct array *a);
+#ifdef PIKE_DEBUG
+void simple_describe_array(struct array *a);
 void describe_index(struct array *a,
 		    int e,
 		    struct processing *p,
 		    int indent);
+#endif
 void describe_array(struct array *a,struct processing *p,int indent);
 PMOD_EXPORT struct array *aggregate_array(INT32 args);
 PMOD_EXPORT struct array *append_array(struct array *a, struct svalue *s);
@@ -162,8 +171,10 @@ PMOD_EXPORT struct array *reverse_array(struct array *a);
 void array_replace(struct array *a,
 		   struct svalue *from,
 		   struct svalue *to);
+#ifdef PIKE_DEBUG
 PMOD_EXPORT void check_array(struct array *a);
 void check_all_arrays(void);
+#endif
 void gc_mark_array_as_referenced(struct array *a);
 void real_gc_cycle_check_array(struct array *a, int weak);
 unsigned gc_touch_all_arrays(void);
@@ -171,10 +182,11 @@ void gc_check_all_arrays(void);
 void gc_mark_all_arrays(void);
 void gc_cycle_check_all_arrays(void);
 void gc_zap_ext_weak_refs_in_arrays(void);
-void gc_free_all_unreferenced_arrays(void);
+size_t gc_free_all_unreferenced_arrays(void);
+#ifdef PIKE_DEBUG
 void debug_dump_type_field(TYPE_FIELD t);
 void debug_dump_array(struct array *a);
-void zap_all_arrays(void);
+#endif
 void count_memory_in_arrays(INT32 *num_, INT32 *size_);
 PMOD_EXPORT struct array *explode_array(struct array *a, struct array *b);
 PMOD_EXPORT struct array *implode_array(struct array *a, struct array *b);
@@ -204,15 +216,18 @@ PMOD_EXPORT struct array *implode_array(struct array *a, struct array *b);
 #define BEGIN_AGGREGATE_ARRAY(estimated_size) do {			\
   struct svalue *base__;						\
   push_array(allocate_array_no_init(0, (estimated_size)));		\
-  base__ = Pike_sp;
+  base__ = Pike_sp;							\
+  base__[-1].u.array->type_field = (BIT_MIXED | BIT_UNFINISHED);
 
 #define DO_AGGREGATE_ARRAY(max_keep_on_stack)				\
   do {									\
     ptrdiff_t diff__ = Pike_sp - base__;				\
     if (diff__ > (max_keep_on_stack)) {					\
       INT32 oldsize__ = base__[-1].u.array->size;			\
-      base__[-1].u.array =						\
-	resize_array(base__[-1].u.array, oldsize__ + diff__);		\
+      ACCEPT_UNFINISHED_TYPE_FIELDS {					\
+	base__[-1].u.array =						\
+	  resize_array(base__[-1].u.array, oldsize__ + diff__);		\
+      } END_ACCEPT_UNFINISHED_TYPE_FIELDS;				\
       /* Unless the user does something, the type field will contain */	\
       /* BIT_MIXED|BIT_UNFINISHED from the allocation above. */		\
       MEMCPY((char *) (ITEM(base__[-1].u.array) + oldsize__),		\
@@ -223,6 +238,10 @@ PMOD_EXPORT struct array *implode_array(struct array *a, struct array *b);
 
 #define END_AGGREGATE_ARRAY						\
   DO_AGGREGATE_ARRAY(0);						\
+  DO_IF_DEBUG(if (Pike_sp[-1].type != T_ARRAY) {			\
+		Pike_fatal("Lost track of aggregated array.\n");	\
+	      });							\
+  array_fix_type_field(Pike_sp[-1].u.array);				\
 } while (0)
 
 
@@ -252,16 +271,50 @@ PMOD_EXPORT struct array *implode_array(struct array *a, struct array *b);
 									 \
   DO_IF_DEBUG(								 \
   if(index_<0 || index_>v_->size)					 \
-    Pike_fatal("Illegal index in low level array set routine.\n");		 \
+    Pike_fatal("Illegal index in low level array set routine.\n");	 \
     )									 \
 									 \
   check_destructed(s_);							 \
   tmp_=ITEM(v_)[index_];						 \
 									 \
-  v_->type_field = (v_->type_field & ~BIT_UNFINISHED) | (1 << s_->type); \
+  v_->type_field |= 1 << s_->type;					 \
   assign_svalue_no_free( ITEM(v_) + index_, s_);			 \
   free_svalue(&tmp_);							 \
 }while(0)
 
+#define array_fix_unfinished_type_field(A) do {				\
+    struct array *a_ = (A);						\
+    if (a_->type_field & BIT_UNFINISHED) {				\
+      DO_IF_DEBUG (array_check_type_field (a_));			\
+      array_fix_type_field (a_);					\
+    }									\
+  } while (0)
+
+#ifdef TRACE_UNFINISHED_TYPE_FIELDS
+/* Note: These macros don't support thread switches. */
+
+extern PMOD_EXPORT int accept_unfinished_type_fields;
+PMOD_EXPORT void dont_accept_unfinished_type_fields (void *orig);
+
+#define ACCEPT_UNFINISHED_TYPE_FIELDS do {				\
+    ONERROR autf_uwp_;							\
+    int orig_autf_ = accept_unfinished_type_fields;			\
+    accept_unfinished_type_fields++;					\
+    SET_ONERROR (autf_uwp_, dont_accept_unfinished_type_fields,		\
+		 (void *) orig_autf_);					\
+    do
+
+#define END_ACCEPT_UNFINISHED_TYPE_FIELDS				\
+    while (0);								\
+    accept_unfinished_type_fields = orig_autf_;				\
+    UNSET_ONERROR (autf_uwp_);						\
+  } while (0)
+
+#else
+
+#define ACCEPT_UNFINISHED_TYPE_FIELDS do
+#define END_ACCEPT_UNFINISHED_TYPE_FIELDS while (0)
+
+#endif
 
 #endif /* ARRAY_H */
