@@ -1,17 +1,27 @@
+/*
+|| This file is part of Pike. For copyright information see COPYRIGHT.
+|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
+|| for more information.
+|| $Id: fdlib.c,v 1.69 2006/07/05 18:07:23 mast Exp $
+*/
+
 #include "global.h"
 #include "fdlib.h"
 #include "pike_error.h"
 #include <math.h>
 
-RCSID("$Id: fdlib.c,v 1.54 2003/09/16 16:45:30 mast Exp $");
+RCSID("$Id: fdlib.c,v 1.69 2006/07/05 18:07:23 mast Exp $");
 
 #ifdef HAVE_WINSOCK_H
 
-#ifdef _REENTRANT
+/* Old versions of the headerfiles don't have this constant... */
+#ifndef INVALID_SET_FILE_POINTER
+#define INVALID_SET_FILE_POINTER ((DWORD)-1)
+#endif
+
 #include "threads.h"
 
 static MUTEX_T fd_mutex;
-#endif
 
 HANDLE da_handle[MAX_OPEN_FILEDESCRIPTORS];
 int fd_type[MAX_OPEN_FILEDESCRIPTORS];
@@ -25,6 +35,38 @@ int first_free_handle;
 #define FDDEBUG(X)
 #endif
 
+#ifdef PIKE_DEBUG
+static int IsValidHandle(HANDLE h)
+{
+  __try {
+    HANDLE ret;
+    if(DuplicateHandle(GetCurrentProcess(),
+			h,
+			GetCurrentProcess(),
+			&ret,
+			0,
+			0,
+			DUPLICATE_SAME_ACCESS))
+    {
+      CloseHandle(ret);
+    }
+  }
+
+  __except (1) {
+    return 0;
+  }
+
+  return 1;
+}
+
+PMOD_EXPORT HANDLE CheckValidHandle(HANDLE h)
+{
+  if(!IsValidHandle(h))
+    Pike_fatal("Invalid handle!\n");
+  return h;
+}
+#endif
+
 PMOD_EXPORT char *debug_fd_info(int fd)
 {
   if(fd<0)
@@ -68,7 +110,7 @@ void fd_init()
   mt_lock(&fd_mutex);
   if(WSAStartup(MAKEWORD(1,1), &wsadata) != 0)
   {
-    fatal("No winsock available.\n");
+    Pike_fatal("No winsock available.\n");
   }
   FDDEBUG(fprintf(stderr,"Using %s\n",wsadata.szDescription));
   
@@ -92,6 +134,10 @@ void fd_exit()
   mt_destroy(&fd_mutex);
 }
 
+/* The following replaces stat(2) in MS libc since it contains a bug
+ * which causes the timestamps for a file to be offset with one hour
+ * whenever a timestamp which has been set during DST is read during
+ * normal time, or vice versa. */
 
 #define ISSEPARATOR(a)  ((a) == '\\' || (a) == '/')
 
@@ -136,7 +182,9 @@ static int IsUncRoot(char *path)
   return 0 ;
 }
 
-static int low_stat (char *file, struct stat *buf)
+static long convert_filetime_to_time_t(FILETIME tmp);
+
+static int low_stat (const char *file, PIKE_STAT_T *buf)
 {
   char            *path;
   int              drive;       /* A: = 1, B: = 2, ... */
@@ -185,86 +233,84 @@ static int low_stat (char *file, struct stat *buf)
     findbuf.nFileSizeHigh = 0;
     findbuf.nFileSizeLow = 0;
     findbuf.cFileName[0] = '\0';
-    
-    buf->st_mtime = __loctotime_t(1980,1,1,0,0,0, -1);
-    buf->st_atime = buf->st_mtime;
-    buf->st_ctime = buf->st_mtime;
+
+    /* Don't have any timestamp info, so just set them to some date.
+     * The following is ridiculously complex, but it's compatible with
+     * the stat function in MS CRT. */
+    {
+      struct tm t;
+      t.tm_year = 1980 - 1900;
+      t.tm_mon  = 0;
+      t.tm_mday = 1;
+      t.tm_hour = 0;
+      t.tm_min  = 0;
+      t.tm_sec  = 0;
+      t.tm_isdst = -1;
+      buf->st_mtime = mktime (&t);
+      buf->st_atime = buf->st_mtime;
+      buf->st_ctime = buf->st_mtime;
+    }
   }
   else
   {
-    SYSTEMTIME SystemTime;
-    SYSTEMTIME UtcSTime;
-    
-    if ( !FileTimeToSystemTime( &findbuf.ftLastWriteTime,
-                                &UtcSTime )                    ||
-         !SystemTimeToTzSpecificLocalTime(NULL, &UtcSTime,
-                                          &SystemTime) )
-    {
-      _dosmaperr( GetLastError() );
-      FindClose( hFind );
-      return( -1 );
-    }
+    buf->st_mtime = convert_filetime_to_time_t (findbuf.ftLastWriteTime);
 
-    buf->st_mtime = __loctotime_t( SystemTime.wYear,
-                                   SystemTime.wMonth,
-                                   SystemTime.wDay,
-                                   SystemTime.wHour,
-                                   SystemTime.wMinute,
-                                   SystemTime.wSecond,
-                                   -1 );
-    
     if ( findbuf.ftLastAccessTime.dwLowDateTime ||
          findbuf.ftLastAccessTime.dwHighDateTime )
     {
-      if ( !FileTimeToSystemTime( &findbuf.ftLastAccessTime,
-                                  &UtcSTime )                    ||
-           !SystemTimeToTzSpecificLocalTime(NULL, &UtcSTime,
-                                            &SystemTime) )
-      {
-        _dosmaperr( GetLastError() );
-        FindClose( hFind );
-        return( -1 );
-      }
-      
-      buf->st_atime = __loctotime_t( SystemTime.wYear,
-                                     SystemTime.wMonth,
-                                     SystemTime.wDay,
-                                     SystemTime.wHour,
-                                     SystemTime.wMinute,
-                                     SystemTime.wSecond,
-                                     -1 );
+      buf->st_atime = convert_filetime_to_time_t (findbuf.ftLastAccessTime);
     } else
       buf->st_atime = buf->st_mtime ;
     
     if ( findbuf.ftCreationTime.dwLowDateTime ||
          findbuf.ftCreationTime.dwHighDateTime )
     {
-      if ( !FileTimeToSystemTime( &findbuf.ftCreationTime,
-                                  &UtcSTime )                    ||
-           !SystemTimeToTzSpecificLocalTime(NULL, &UtcSTime,
-                                            &SystemTime) )
-      {
-        _dosmaperr( GetLastError() );
-        FindClose( hFind );
-        return( -1 );
-      }
-      
-      buf->st_ctime = __loctotime_t( SystemTime.wYear,
-                                     SystemTime.wMonth,
-                                     SystemTime.wDay,
-                                     SystemTime.wHour,
-                                     SystemTime.wMinute,
-                                     SystemTime.wSecond,
-                                     -1 );
+      buf->st_atime = convert_filetime_to_time_t (findbuf.ftCreationTime);
     } else
       buf->st_ctime = buf->st_mtime ;
     
     FindClose(hFind);
   }
-  
-  buf->st_mode = __dtoxmode(findbuf.dwFileAttributes, file);
+
+  if (findbuf.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
+    /* Always label a directory as executable. Note that this also
+     * catches special locations like root dirs and network share
+     * roots. */
+    buf->st_mode = S_IFDIR | 0111;
+  else {
+    const char *p;
+    buf->st_mode = S_IFREG;
+
+    /* Look at the extension to see if a file appears to be executable. */
+    if ((p = strrchr (file, '.')) && strlen (p) == 4) {
+      char ext[4];
+      ext[0] = tolower (p[1]);
+      ext[1] = tolower (p[2]);
+      ext[2] = tolower (p[3]);
+      ext[3] = 0;
+      if (!strcmp (ext, "exe") ||
+	  !strcmp (ext, "cmd") ||
+	  !strcmp (ext, "bat") ||
+	  !strcmp (ext, "com"))
+	buf->st_mode |= 0111;	/* Execute perm for all. */
+    }
+  }
+
+  /* The file is read/write unless the read only flag is set. */
+  if (findbuf.dwFileAttributes & FILE_ATTRIBUTE_READONLY)
+    buf->st_mode |= 0444;	/* Read perm for all. */
+  else
+    buf->st_mode |= 0666;	/* Read and write perm for all. */
+
   buf->st_nlink = 1;
-  buf->st_size = findbuf.nFileSizeLow;
+#ifdef INT64
+  buf->st_size = ((INT64) findbuf.nFileSizeHigh << 32) + findbuf.nFileSizeLow;
+#else
+  if (findbuf.nFileSizeHigh)
+    buf->st_size = MAXDWORD;
+  else
+    buf->st_size = findbuf.nFileSizeLow;
+#endif
   
   buf->st_uid = buf->st_gid = buf->st_ino = 0; /* unused entries */
   buf->st_rdev = buf->st_dev = (_dev_t)(drive - 1); /* A=0, B=1, ... */
@@ -272,7 +318,7 @@ static int low_stat (char *file, struct stat *buf)
   return(0);
 }
 
-int debug_fd_stat(char *file, struct stat *buf)
+int debug_fd_stat(const char *file, PIKE_STAT_T *buf)
 {
   ptrdiff_t l = strlen(file);
   char fname[MAX_PATH];
@@ -294,7 +340,7 @@ int debug_fd_stat(char *file, struct stat *buf)
   return low_stat(file, buf);
 }
 
-PMOD_EXPORT FD debug_fd_open(char *file, int open_mode, int create_mode)
+PMOD_EXPORT FD debug_fd_open(const char *file, int open_mode, int create_mode)
 {
   HANDLE x;
   FD fd;
@@ -751,9 +797,9 @@ PMOD_EXPORT ptrdiff_t debug_fd_read(FD fd, void *to, ptrdiff_t len)
   }
 }
 
-PMOD_EXPORT ptrdiff_t debug_fd_lseek(FD fd, ptrdiff_t pos, int where)
+PMOD_EXPORT PIKE_OFF_T debug_fd_lseek(FD fd, PIKE_OFF_T pos, int where)
 {
-  ptrdiff_t ret;
+  PIKE_OFF_T ret;
   HANDLE h;
 
   mt_lock(&fd_mutex);
@@ -774,22 +820,38 @@ PMOD_EXPORT ptrdiff_t debug_fd_lseek(FD fd, ptrdiff_t pos, int where)
   h = da_handle[fd];
   mt_unlock(&fd_mutex);
 
-  ret = SetFilePointer(h,
-		       DO_NOT_WARN((LONG)pos),
-		       0, where);
-  if(!~ret)
-  {
-    errno=GetLastError();
-    return -1;
+#ifdef INT64
+  if (pos >= ((INT64) 1 << 32)) {
+    LONG high = DO_NOT_WARN ((LONG) (pos >> 32));
+    DWORD err;
+    pos &= ((INT64) 1 << 32) - 1;
+    ret = SetFilePointer (h, DO_NOT_WARN ((LONG) pos), &high, where);
+    if (ret == INVALID_SET_FILE_POINTER &&
+	(err = GetLastError()) != NO_ERROR) {
+      errno = err;
+      return -1;
+    }
+    ret += (INT64) high << 32;
   }
+  else
+#endif
+  {
+    ret = SetFilePointer (h, DO_NOT_WARN ((LONG) pos), NULL, where);
+    if(ret == INVALID_SET_FILE_POINTER)
+    {
+      errno=GetLastError();
+      return -1;
+    }
+  }
+
   return ret;
 }
 
-PMOD_EXPORT int debug_fd_ftruncate(FD fd, ptrdiff_t len)
+PMOD_EXPORT int debug_fd_ftruncate(FD fd, PIKE_OFF_T len)
 {
-  ptrdiff_t ret;
   HANDLE h;
-  LONG oldfp_lo, oldfp_hi;
+  LONG oldfp_lo, oldfp_hi, len_hi;
+  DWORD err;
 
   mt_lock(&fd_mutex);
   if(fd_type[fd]!=FD_FILE)
@@ -804,23 +866,43 @@ PMOD_EXPORT int debug_fd_ftruncate(FD fd, ptrdiff_t len)
   oldfp_hi = 0;
   oldfp_lo = SetFilePointer(h, 0, &oldfp_hi, FILE_CURRENT);
   if(!~oldfp_lo) {
-    errno=GetLastError();
-    if(errno != NO_ERROR)
+    err = GetLastError();
+    if(err != NO_ERROR) {
+      errno = err;
       return -1;
+    }
   }
-  if(!~SetFilePointer(h,
-		      DO_NOT_WARN((LONG)len),
-		      NULL, FILE_BEGIN) ||
-     !SetEndOfFile(h)) {
+
+#ifdef INT64
+  len_hi = DO_NOT_WARN ((LONG) (len >> 32));
+  len &= ((INT64) 1 << 32) - 1;
+#else
+  len_hi = 0;
+#endif
+
+  if (SetFilePointer (h, DO_NOT_WARN ((LONG) len), &len_hi, FILE_BEGIN) ==
+      INVALID_SET_FILE_POINTER &&
+      (err = GetLastError()) != NO_ERROR) {
+    SetFilePointer(h, oldfp_lo, &oldfp_hi, FILE_BEGIN);
+    errno = err;
+    return -1;
+  }
+
+  if(!SetEndOfFile(h)) {
     errno=GetLastError();
     SetFilePointer(h, oldfp_lo, &oldfp_hi, FILE_BEGIN);
     return -1;
   }
-  if(!~SetFilePointer(h, oldfp_lo, &oldfp_hi, FILE_BEGIN)) {
-    errno=GetLastError();
-    if(errno != NO_ERROR)
-      return -1;
-  }
+
+  if (oldfp_hi < len_hi || oldfp_hi == len_hi && oldfp_lo < len)
+    if(!~SetFilePointer(h, oldfp_lo, &oldfp_hi, FILE_BEGIN)) {
+      err = GetLastError();
+      if(err != NO_ERROR) {
+	errno = err;
+	return -1;
+      }
+    }
+
   return 0;
 }
 
@@ -878,17 +960,28 @@ PMOD_EXPORT int debug_fd_flock(FD fd, int oper)
 static long convert_filetime_to_time_t(FILETIME tmp)
 {
   double t;
+  /* FILETIME is in 100ns since Jan 01, 1601 00:00 UTC.
+   *
+   * Offset to Jan 01, 1970 is thus 0x019db1ded53e8000 * 100ns.
+   */
+  if (tmp.dwLowDateTime < 0xd53e8000UL) {
+    tmp.dwHighDateTime -= 0x019db1dfUL;	/* Note: Carry! */
+    tmp.dwLowDateTime += 0x2ac18000UL;	/* Note: 2-compl */
+  } else {
+    tmp.dwHighDateTime -= 0x019db1deUL;
+    tmp.dwLowDateTime -= 0xd53e8000UL;
+  }
   t=tmp.dwHighDateTime * pow(2.0,32.0) + (double)tmp.dwLowDateTime;
+
+  /* 1s == 10000000 * 100ns. */
   t/=10000000.0;
-  t-=11644473600.0;
   return DO_NOT_WARN((long)floor(t));
 }
 
-PMOD_EXPORT int debug_fd_fstat(FD fd, struct stat *s)
+PMOD_EXPORT int debug_fd_fstat(FD fd, PIKE_STAT_T *s)
 {
-  DWORD x;
-
   FILETIME c,a,m;
+
   mt_lock(&fd_mutex);
   FDDEBUG(fprintf(stderr, "fstat on %d (%ld)\n",
 		  fd, PTRDIFF_T_TO_LONG((ptrdiff_t)da_handle[fd])));
@@ -899,7 +992,7 @@ PMOD_EXPORT int debug_fd_fstat(FD fd, struct stat *s)
     return -1;
   }
 
-  MEMSET(s, 0, sizeof(struct stat));
+  MEMSET(s, 0, sizeof(PIKE_STAT_T));
   s->st_nlink=1;
 
   switch(fd_type[fd])
@@ -914,16 +1007,26 @@ PMOD_EXPORT int debug_fd_fstat(FD fd, struct stat *s)
 	default:
 	case FILE_TYPE_UNKNOWN: s->st_mode=0;        break;
 	case FILE_TYPE_DISK:
-	  s->st_mode=S_IFREG; 
-	  s->st_size=GetFileSize(da_handle[fd],&x);
-	  if(x)
+	  s->st_mode=S_IFREG;
 	  {
-	    s->st_size=0x7fffffff;
+	    DWORD high, err;
+	    s->st_size=GetFileSize(da_handle[fd],&high);
+	    if (s->st_size == INVALID_FILE_SIZE &&
+		(err = GetLastError()) != NO_ERROR) {
+	      errno = err;
+	      mt_unlock(&fd_mutex);
+	      return -1;
+	    }
+#ifdef INT64
+	    s->st_size += (INT64) high << 32;
+#else
+	    if (high) s->st_size = MAXDWORD;
+#endif
 	  }
 	  if(!GetFileTime(da_handle[fd], &c, &a, &m))
 	  {
 	    errno=GetLastError();
-            mt_unlock(&fd_mutex);
+	    mt_unlock(&fd_mutex);
 	    return -1;
 	  }
 	  s->st_ctime=convert_filetime_to_time_t(c);
@@ -1029,17 +1132,19 @@ PMOD_EXPORT FD debug_fd_dup(FD from)
 {
   FD fd;
   HANDLE x,p=GetCurrentProcess();
+
+  mt_lock(&fd_mutex);
 #ifdef DEBUG
   if(fd_type[from]>=FD_NO_MORE_FREE)
-    fatal("fd_dup() on file which is not open!\n");
+    Pike_fatal("fd_dup() on file which is not open!\n");
 #endif
   if(!DuplicateHandle(p,da_handle[from],p,&x,0,0,DUPLICATE_SAME_ACCESS))
   {
     errno=GetLastError();
+    mt_unlock(&fd_mutex);
     return -1;
   }
 
-  mt_lock(&fd_mutex);
   fd=first_free_handle;
   first_free_handle=fd_type[fd];
   fd_type[fd]=fd_type[from];
@@ -1054,13 +1159,15 @@ PMOD_EXPORT FD debug_fd_dup(FD from)
 PMOD_EXPORT FD debug_fd_dup2(FD from, FD to)
 {
   HANDLE x,p=GetCurrentProcess();
+
+  mt_lock(&fd_mutex);
   if(!DuplicateHandle(p,da_handle[from],p,&x,0,0,DUPLICATE_SAME_ACCESS))
   {
     errno=GetLastError();
+    mt_unlock(&fd_mutex);
     return -1;
   }
 
-  mt_lock(&fd_mutex);
   if(fd_type[to] < FD_NO_MORE_FREE)
   {
     if(!CloseHandle(da_handle[to]))
@@ -1175,7 +1282,7 @@ void fd_mapper_set(struct fd_mapper *x, FD fd, void *data)
     x->size*=2;
     x->data=(void **)realloc((char *)x->data, x->size*sizeof(void *));
     if(!x->data)
-      fatal("Out of memory.\n");
+      Pike_fatal("Out of memory.\n");
     x->data=nd;
   }
   x->data[fd]=data;
@@ -1501,7 +1608,7 @@ void fd_waitor_remove_customer(fd_waitor *x, FD customer)
 
   CloseHandle(x->customers[pos]);
   
-  fd_mapper_store(customer, x->fd_to_pos_key, (void *)0);
+  fd_mapper_store(customer, x->fd_to_pos_key, NULL);
   x->occupied--;
   if(x->occupied != pos)
   {
