@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: interpret.h,v 1.125 2004/05/01 16:46:08 marcus Exp $
+|| $Id: interpret.h,v 1.152 2004/05/21 16:26:29 grubba Exp $
 */
 
 #ifndef INTERPRET_H
@@ -11,7 +11,7 @@
 #include "global.h"
 #include "program.h"
 #include "pike_error.h"
-#include "bignum.h"
+#include "object.h"
 
 struct Pike_interpreter {
   /* Swapped variables */
@@ -24,7 +24,7 @@ struct Pike_interpreter {
   int mark_stack_malloced;
   JMP_BUF *recoveries;
 #ifdef PIKE_THREADS
-  struct object * thread_id;
+  struct thread_state *thread_state;
 #endif
   char *stack_top;
   DO_IF_SECURITY(struct object *current_creds;)
@@ -40,9 +40,7 @@ struct Pike_interpreter {
   char *stack_bottom;
 #endif
 
-#ifdef THREAD_TRACE
-  int t_flag;
-#endif /* THREAD_TRACE */
+  int trace_level;
 };
 
 #ifndef STRUCT_FRAME_DECLARED
@@ -194,41 +192,248 @@ PMOD_EXPORT extern const char Pike_check_c_stack_errmsg[];
 
 PMOD_EXPORT extern const char msg_pop_neg[];
 #define pop_n_elems(X)							\
- do { ptrdiff_t x_=(X); if(x_) { 					\
-   check__positive(x_, (msg_pop_neg, x_));				\
-   Pike_sp -= x_; debug_check_stack();					\
-   free_mixed_svalues(Pike_sp, x_);					\
- } } while (0)
+ do {									\
+   ptrdiff_t x_=(X);							\
+   if(x_) {								\
+     struct svalue *_sp_;						\
+     check__positive(x_, (msg_pop_neg, x_));				\
+     _sp_ = Pike_sp = Pike_sp - x_;					\
+     debug_check_stack();						\
+     free_mixed_svalues(_sp_, x_);					\
+   }									\
+ } while (0)
 
-#define stack_pop_n_elems_keep_top(X) \
- do { struct svalue s=Pike_sp[-1]; Pike_sp[-1]=Pike_sp[-1-(X)]; Pike_sp[-1-(X)]=s; \
-      pop_n_elems(X); } while (0)
+/* This pops a number of arguments from the stack but keeps the top
+ * element on top. Used for popping the arguments while keeping the
+ * return value.
+ */
+#define stack_unlink(X) do {						\
+    ptrdiff_t x2_ = (X);						\
+    if (x2_) {								\
+      struct svalue *_sp_ = --Pike_sp;					\
+      free_svalue (_sp_ - x2_);						\
+      move_svalue (_sp_ - x2_, _sp_);					\
+      pop_n_elems (x2_ - 1);						\
+    }									\
+  }while(0)
 
-#define push_program(P) do{ struct program *_=(P); debug_malloc_touch(_); Pike_sp->u.program=_; Pike_sp++->type=PIKE_T_PROGRAM; }while(0)
-#define push_int(I) do{ INT_TYPE _=(I); Pike_sp->u.integer=_;Pike_sp->type=PIKE_T_INT;Pike_sp++->subtype=NUMBER_NUMBER; }while(0)
-#define push_undefined() do{  Pike_sp->u.integer=0; Pike_sp->type=PIKE_T_INT; Pike_sp++->subtype=NUMBER_UNDEFINED; }while(0)
-#define push_mapping(M) do{ struct mapping *_=(M); debug_malloc_touch(_); Pike_sp->u.mapping=_; Pike_sp++->type=PIKE_T_MAPPING; }while(0)
-#define push_array(A) do{ struct array *_=(A); debug_malloc_touch(_); Pike_sp->u.array=_ ;Pike_sp++->type=PIKE_T_ARRAY; }while(0)
-#define push_multiset(L) do{ struct multiset *_=(L); debug_malloc_touch(_); Pike_sp->u.multiset=_; Pike_sp++->type=PIKE_T_MULTISET; }while(0)
-#define push_string(S) do{ struct pike_string *_=(S); debug_malloc_touch(_); Pike_sp->subtype=0; Pike_sp->u.string=_; Pike_sp++->type=PIKE_T_STRING; }while(0)
-#define push_type_value(S) do{ struct pike_type *_=(S); debug_malloc_touch(_); Pike_sp->u.type=_; Pike_sp++->type=PIKE_T_TYPE; }while(0)
-#define push_object(O) do{ struct object  *_=(O); debug_malloc_touch(_); Pike_sp->u.object=_; Pike_sp++->type=PIKE_T_OBJECT; }while(0)
-#define push_float(F) do{ FLOAT_TYPE _=(F); Pike_sp->u.float_number=_; Pike_sp++->type=PIKE_T_FLOAT; }while(0)
-#define push_text(T) push_string(make_shared_string((T)))
-#define push_constant_text(T) do{ Pike_sp->subtype=0; MAKE_CONSTANT_SHARED_STRING(Pike_sp->u.string,T); Pike_sp++->type=PIKE_T_STRING; }while(0)
+#define stack_pop_n_elems_keep_top(X) stack_unlink(X)
 
-#define ref_push_program(P) do{ struct program *_=(P); debug_malloc_touch(_); _->refs++; Pike_sp->u.program=_; Pike_sp++->type=PIKE_T_PROGRAM; }while(0)
-#define ref_push_mapping(M) do{ struct mapping *_=(M); debug_malloc_touch(_); _->refs++; Pike_sp->u.mapping=_; Pike_sp++->type=PIKE_T_MAPPING; }while(0)
-#define ref_push_array(A) do{ struct array *_=(A); debug_malloc_touch(_); _->refs++; Pike_sp->u.array=_ ;Pike_sp++->type=PIKE_T_ARRAY; }while(0)
-#define ref_push_multiset(L) do{ struct multiset *_=(L); debug_malloc_touch(_); _->refs++; Pike_sp->u.multiset=_; Pike_sp++->type=PIKE_T_MULTISET; }while(0)
-#define ref_push_string(S) do{ struct pike_string *_=(S); debug_malloc_touch(_); _->refs++; Pike_sp->subtype=0; Pike_sp->u.string=_; Pike_sp++->type=PIKE_T_STRING; }while(0)
-#define ref_push_type_value(S) do{ struct pike_type *_=(S); debug_malloc_touch(_); _->refs++; Pike_sp->u.type=_; Pike_sp++->type=PIKE_T_TYPE; }while(0)
-#define ref_push_object(O) do{ struct object  *_=(O); debug_malloc_touch(_); _->refs++; Pike_sp->u.object=_; Pike_sp++->type=PIKE_T_OBJECT; }while(0)
+#define stack_pop_keep_top() do {					\
+    struct svalue *_sp_ = --Pike_sp;					\
+    free_svalue (_sp_ - 1);						\
+    move_svalue (_sp_ - 1, _sp_);					\
+    debug_check_stack();						\
+  } while (0)
 
-#define push_svalue(S) do { const struct svalue *_=(S); assign_svalue_no_free(Pike_sp,_); Pike_sp++; }while(0)
+#define stack_pop_2_elems_keep_top() do {				\
+    struct svalue *_sp_ = Pike_sp = Pike_sp - 2;			\
+    free_svalue (_sp_ - 1);						\
+    free_svalue (_sp_);							\
+    move_svalue (_sp_ - 1, _sp_ + 1);					\
+    debug_check_stack();						\
+  } while (0)
+
+#define stack_pop_to_no_free(X) move_svalue(X, --Pike_sp)
+
+#define stack_pop_to(X) do {						\
+    struct svalue *_=(X);						\
+    free_svalue(_);							\
+    stack_pop_to_no_free(_);						\
+  }while(0)
+
+#define push_program(P) do{						\
+    struct program *_=(P);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    debug_malloc_touch(_);						\
+    _sp_->u.program=_;							\
+    _sp_++->type=PIKE_T_PROGRAM;					\
+  }while(0)
+
+#define push_int(I) do{							\
+    INT_TYPE _=(I);							\
+    struct svalue *_sp_ = Pike_sp++;					\
+    _sp_->u.integer=_;							\
+    _sp_->type=PIKE_T_INT;						\
+    _sp_->subtype=NUMBER_NUMBER;					\
+  }while(0)
+
+#define push_undefined() do{						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    _sp_->u.integer=0;							\
+    _sp_->type=PIKE_T_INT;						\
+    _sp_->subtype=NUMBER_UNDEFINED;					\
+  }while(0)
+
+#define push_mapping(M) do{						\
+    struct mapping *_=(M);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    debug_malloc_touch(_);						\
+    _sp_->u.mapping=_;							\
+    _sp_->type=PIKE_T_MAPPING;						\
+  }while(0)
+
+#define push_array(A) do{						\
+    struct array *_=(A);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    debug_malloc_touch(_);						\
+    _sp_->u.array=_ ;							\
+    _sp_->type=PIKE_T_ARRAY;						\
+  }while(0)
+
+#define push_multiset(L) do{						\
+    struct multiset *_=(L);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    debug_malloc_touch(_);						\
+    _sp_->u.multiset=_;							\
+    _sp_->type=PIKE_T_MULTISET;						\
+  }while(0)
+
+#define push_string(S) do {						\
+    struct pike_string *_=(S);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    debug_malloc_touch(_);						\
+    DO_IF_DEBUG(if(_->size_shift & ~3) {				\
+		  Pike_fatal("Pushing string with bad shift: %d\n",	\
+			     _->size_shift);				\
+		});							\
+    _sp_->subtype=0;							\
+    _sp_->u.string=_;							\
+    _sp_->type=PIKE_T_STRING;						\
+  }while(0)
+
+#define push_type_value(S) do{						\
+    struct pike_type *_=(S);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    debug_malloc_touch(_);						\
+    _sp_->u.type=_;							\
+    _sp_->type=PIKE_T_TYPE;						\
+  }while(0)
+
+#define push_object(O) do {						\
+    struct object *_=(O);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    debug_malloc_touch(_);						\
+    _sp_->u.object=_;							\
+    _sp_->type=PIKE_T_OBJECT;						\
+  }while(0)
+
+#define push_float(F) do{						\
+    FLOAT_TYPE _=(F);							\
+    struct svalue *_sp_ = Pike_sp++;					\
+    _sp_->u.float_number=_;						\
+    _sp_->type=PIKE_T_FLOAT;						\
+  }while(0)
+
+#define push_text(T) do {						\
+    const char *_ = (T);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    _sp_->subtype=0;							\
+    _sp_->u.string=make_shared_binary_string(_,strlen(_));		\
+    debug_malloc_touch(_sp_->u.string);					\
+    _sp_->type=PIKE_T_STRING;						\
+  }while(0)
+
+#define push_constant_text(T) do{					\
+    struct svalue *_sp_ = Pike_sp++;					\
+    _sp_->subtype=0;							\
+    REF_MAKE_CONST_STRING(_sp_->u.string,T);				\
+    _sp_->type=PIKE_T_STRING;						\
+  }while(0)
+
+#define push_function(OBJ, FUN) do {					\
+    struct object *_=(OBJ);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    debug_malloc_touch(_);						\
+    _sp_->u.object=_;							\
+    _sp_->subtype=(FUN);						\
+    _sp_->type=PIKE_T_FUNCTION;						\
+  } while (0)
+
+#define ref_push_program(P) do{						\
+    struct program *_=(P);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    add_ref(_);								\
+    _sp_->u.program=_;							\
+    _sp_->type=PIKE_T_PROGRAM;						\
+  }while(0)
+
+#define ref_push_mapping(M) do{						\
+    struct mapping *_=(M);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    add_ref(_);								\
+    _sp_->u.mapping=_;							\
+    _sp_->type=PIKE_T_MAPPING;						\
+  }while(0)
+
+#define ref_push_array(A) do{						\
+    struct array *_=(A);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    add_ref(_);								\
+    _sp_->u.array=_ ;							\
+    _sp_->type=PIKE_T_ARRAY;						\
+  }while(0)
+
+#define ref_push_multiset(L) do{					\
+    struct multiset *_=(L);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    add_ref(_);								\
+    _sp_->u.multiset=_;							\
+    _sp_->type=PIKE_T_MULTISET;						\
+  }while(0)
+
+#define ref_push_string(S) do{						\
+    struct pike_string *_=(S);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    DO_IF_DEBUG(if(_->size_shift & ~3) {				\
+		  Pike_fatal("Pushing string with bad shift: %d\n",	\
+			     _->size_shift);				\
+		});							\
+    add_ref(_);								\
+    _sp_->subtype=0;							\
+    _sp_->u.string=_;							\
+    _sp_->type=PIKE_T_STRING;						\
+  }while(0)
+
+#define ref_push_type_value(S) do{					\
+    struct pike_type *_=(S);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    add_ref(_);								\
+    _sp_->u.type=_;							\
+    _sp_->type=PIKE_T_TYPE;						\
+  }while(0)
+
+#define ref_push_object(O) do{						\
+    struct object  *_=(O);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    add_ref(_);								\
+    _sp_->u.object=_;							\
+    _sp_->type=PIKE_T_OBJECT;						\
+  }while(0)
+
+#define ref_push_function(OBJ, FUN) do {				\
+    struct object *_=(OBJ);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    add_ref(_);								\
+    _sp_->u.object=_;							\
+    _sp_->subtype=(FUN);						\
+    _sp_->type=PIKE_T_FUNCTION;						\
+  } while (0)
+
+#define push_svalue(S) do {						\
+    const struct svalue *_=(S);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    assign_svalue_no_free(_sp_,_);					\
+  }while(0)
 
 #define stack_dup() push_svalue(Pike_sp-1)
-#define stack_swap() do { struct svalue _=Pike_sp[-1]; Pike_sp[-1]=Pike_sp[-2]; Pike_sp[-2]=_; } while(0)
+
+#define stack_swap() do {						\
+    struct svalue *_sp_ = Pike_sp;					\
+    struct svalue _=_sp_[-1];						\
+    _sp_[-1]=_sp_[-2];							\
+    _sp_[-2]=_;								\
+  } while(0)
 
 #define push_zeroes(N) do{			\
     struct svalue *s_=Pike_sp;			\
@@ -254,39 +459,62 @@ PMOD_EXPORT extern const char msg_pop_neg[];
     Pike_sp=s_;					\
 }while(0)
 
-#define stack_pop_to_no_free(X) move_svalue(X, --Pike_sp)
-#define stack_pop_to(X) do { struct svalue *_=(X); free_svalue(_); stack_pop_to_no_free(_); }while(0)
-
-/* This pops a number of arguments from the stack but keeps the top
- * element on top. Used for popping the arguments while keeping the
- * return value.
- */
-#define stack_unlink(X) do { if(X) { free_svalue(Pike_sp-(X)-1); Pike_sp[-(X)-1]=Pike_sp[-1]; Pike_sp--; pop_n_elems(X-1); } }while(0)
-
-#define free_pike_frame(F) do{ struct pike_frame *f_=(F); debug_malloc_touch(f_); if(!--f_->refs) really_free_pike_frame(f_); }while(0)
+#define free_pike_frame(F) do{ struct pike_frame *f_=(F); if(!sub_ref(f_)) really_free_pike_frame(f_); }while(0)
 
 /* A scope is any frame which may have malloced locals */
-#define free_pike_scope(F) do{ struct pike_frame *f_=(F); debug_malloc_touch(f_); if(!--f_->refs) really_free_pike_scope(f_); }while(0)
+#define free_pike_scope(F) do{ struct pike_frame *f_=(F); if(!sub_ref(f_)) really_free_pike_scope(f_); }while(0)
 
 #define POP_PIKE_FRAME() do {						\
-  struct pike_frame *tmp_=Pike_fp->next;					\
-  if(!--Pike_fp->refs)							\
+  struct pike_frame *_fp_ = Pike_fp;					\
+  struct pike_frame *tmp_=_fp_->next;					\
+  DO_IF_PROFILING({							\
+      /* Time spent in this frame + children. */			\
+      cpu_time_t time_passed =						\
+	gethrtime() - Pike_interpreter.time_base;			\
+      /* Time spent in children to this frame. */			\
+      cpu_time_t time_in_children;					\
+      /* Time spent in just this frame. */				\
+      cpu_time_t self_time;						\
+      struct identifier *function;					\
+      time_passed -= _fp_->start_time;					\
+      time_in_children =						\
+	Pike_interpreter.accounted_time - _fp_->children_base;		\
+      self_time = time_passed - time_in_children;			\
+      Pike_interpreter.accounted_time += self_time;			\
+      /* FIXME: Can context.prog be NULL? */				\
+      function = _fp_->context.prog->identifiers + _fp_->ident;		\
+      /* function->total_time =						\
+	 Pike_fp->self_time_base + time_passed; */			\
+      function->total_time += time_passed;				\
+      function->self_time += self_time;					\
+    });									\
+  if(!sub_ref(_fp_))							\
   {									\
-    really_free_pike_frame(Pike_fp);						\
+    really_free_pike_frame(_fp_);					\
   }else{								\
-    DO_IF_DEBUG(if( Pike_fp->locals + Pike_fp->num_locals > Pike_sp || Pike_sp < Pike_fp->expendible) Pike_fatal("Stack failure in POP_PIKE_FRAME %p+%d=%p %p %p!\n",Pike_fp->locals,Pike_fp->num_locals,Pike_fp->locals+Pike_fp->num_locals,Pike_sp,Pike_fp->expendible));                      \
-    debug_malloc_touch(Pike_fp); \
-    if(Pike_fp->num_locals)							\
+    ptrdiff_t num_expendible = _fp_->expendible - _fp_->locals;		\
+    DO_IF_DEBUG(							\
+      if( (_fp_->locals + _fp_->num_locals > Pike_sp) ||		\
+	  (Pike_sp < _fp_->expendible) ||				\
+	  (num_expendible < 0) || (num_expendible > _fp_->num_locals))	\
+	Pike_fatal("Stack failure in POP_PIKE_FRAME %p+%d=%p %p %p!\n",	\
+		   _fp_->locals, _fp_->num_locals,			\
+		   _fp_->locals+_fp_->num_locals,			\
+		   Pike_sp,_fp_->expendible));				\
+    debug_malloc_touch(_fp_);						\
+    if(num_expendible)							\
     {									\
       struct svalue *s=(struct svalue *)xalloc(sizeof(struct svalue)*	\
-					       Pike_fp->num_locals);		\
-      assign_svalues_no_free(s,Pike_fp->locals,Pike_fp->num_locals,BIT_MIXED);	\
-      Pike_fp->locals=s;							\
-      Pike_fp->flags|=PIKE_FRAME_MALLOCED_LOCALS;				\
+					       num_expendible);		\
+      _fp_->num_locals = num_expendible;				\
+      assign_svalues_no_free(s, _fp_->locals, num_expendible,		\
+			     BIT_MIXED);				\
+      _fp_->locals=s;							\
+      _fp_->flags|=PIKE_FRAME_MALLOCED_LOCALS;				\
     }else{								\
-      Pike_fp->locals=0;							\
+      _fp_->locals=0;							\
     }									\
-    Pike_fp->next=0;								\
+    _fp_->next=0;							\
   }									\
   Pike_fp=tmp_;								\
  }while(0)
@@ -358,7 +586,7 @@ PMOD_EXPORT extern unsigned long evaluator_callback_calls;
 
 #define check_threads_etc() do { \
   DO_IF_INTERNAL_PROFILING (evaluator_callback_calls++); \
-  call_callback(& evaluator_callbacks, (void *)0); \
+  call_callback(& evaluator_callbacks, NULL); \
 }while(0) 
 
 #ifdef PIKE_DEBUG
@@ -374,6 +602,8 @@ PMOD_EXPORT extern unsigned long evaluator_callback_calls;
 /* Prototypes begin here */
 void push_sp_mark(void);
 ptrdiff_t pop_sp_mark(void);
+void gc_mark_stack_external (struct pike_frame *frame,
+			     struct svalue *stack_p, struct svalue *stack);
 PMOD_EXPORT void init_interpreter(void);
 void lvalue_to_svalue_no_free(struct svalue *to,struct svalue *lval);
 PMOD_EXPORT void assign_lvalue(struct svalue *lval,struct svalue *from);
@@ -394,7 +624,7 @@ void simple_debug_instr_prologue_0 (PIKE_INSTR_T instr);
 void simple_debug_instr_prologue_1 (PIKE_INSTR_T instr, INT32 arg);
 void simple_debug_instr_prologue_2 (PIKE_INSTR_T instr, INT32 arg1, INT32 arg2);
 #endif
-#endif
+#endif	/* PIKE_USE_MACHINE_CODE */
 
 PMOD_EXPORT void find_external_context(struct external_variable_context *loc,
 				       int arg2);
@@ -429,6 +659,7 @@ PMOD_EXPORT void apply_shared(struct object *o,
 		  int args);
 PMOD_EXPORT void apply(struct object *o, const char *fun, int args);
 PMOD_EXPORT void apply_svalue(struct svalue *s, INT32 args);
+PMOD_EXPORT void apply_external(int depth, int fun, INT32 args);
 PMOD_EXPORT void slow_check_stack(void);
 PMOD_EXPORT void custom_check_stack(ptrdiff_t amount, const char *fmt, ...)
   ATTRIBUTE((format (printf, 2, 3)));
@@ -457,6 +688,11 @@ static inline void strict_apply_svalue(struct svalue *sval, INT32 args)
   mega_apply(APPLY_SVALUE, (ARGS), (void*)(SVAL),0)
 #endif /* __ECL */
 
+#define apply_current(FUN, ARGS)			\
+  apply_low(Pike_fp->current_object,			\
+	    (FUN) + Pike_fp->context.identifier_level,	\
+	    (ARGS))
+
 PMOD_EXPORT extern int d_flag; /* really in main.c */
 
 PMOD_EXPORT extern int Pike_stack_size;
@@ -479,9 +715,6 @@ PMOD_EXPORT extern struct Pike_interpreter Pike_interpreter;
 #define Pike_sp Pike_interpreter.stack_pointer
 #define Pike_fp Pike_interpreter.frame_pointer
 #define Pike_mark_sp Pike_interpreter.mark_stack_pointer
-#ifdef PIKE_THREADS
-#define Pike_thread_id Pike_interpreter.thread_id
-#endif
 
 
 #define CURRENT_STORAGE (dmalloc_touch(struct pike_frame *,Pike_fp)->current_storage)
