@@ -2,45 +2,26 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: xbm.c,v 1.18 2002/10/21 17:06:15 marcus Exp $
+|| $Id: xbm.c,v 1.23 2004/03/06 00:07:00 nilsson Exp $
 */
 
 #define NO_PIKE_SHORTHAND
 
 #include "global.h"
-RCSID("$Id: xbm.c,v 1.18 2002/10/21 17:06:15 marcus Exp $");
+RCSID("$Id: xbm.c,v 1.23 2004/03/06 00:07:00 nilsson Exp $");
 
 #include "image_machine.h"
 
 #include "interpret.h"
 #include "svalue.h"
-#include "pike_macros.h"
 #include "object.h"
 #include "program.h"
-#include "array.h"
 #include "pike_error.h"
-#include "constants.h"
 #include "mapping.h"
 #include "stralloc.h"
-#include "multiset.h"
-#include "pike_types.h"
-#include "rusage.h"
 #include "operators.h"
-#include "fsort.h"
-#include "callback.h"
-#include "backend.h"
-#include "main.h"
-#include "pike_memory.h"
 #include "threads.h"
-#include "time_stuff.h"
-#include "version.h"
-#include "encode.h"
 #include "module_support.h"
-#include "module.h"
-#include "opcodes.h"
-#include "cyclic.h"
-#include "signal_handler.h"
-#include "security.h"
 #include "builtin_functions.h"
 
 #include "image.h"
@@ -188,11 +169,11 @@ static struct pike_string *save_xbm( struct image *i, struct pike_string *name )
 
   initialize_buf(&buf);
   ccat( "#define ");  cname();  ccat( "_width " );
-  sprintf( size, "%d\n", i->xsize );
+  sprintf( size, "%"PRINTPIKEINT"d\n", i->xsize );
   low_my_binary_strcat( size, strlen(size), &buf );
 
   ccat( "#define ");  cname();  ccat( "_height " );
-  sprintf( size, "%d\n", i->ysize );
+  sprintf( size, "%"PRINTPIKEINT"d\n", i->ysize );
   low_my_binary_strcat( size, strlen(size), &buf );
 
   ccat( "static char " );  cname();  ccat( "_bits[] = {\n" );
