@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: oracle.c,v 1.84 2004/10/12 17:01:04 mast Exp $
+|| $Id: oracle.c,v 1.86 2005/08/17 09:30:04 grubba Exp $
 */
 
 /*
@@ -53,7 +53,7 @@
 
 #include <math.h>
 
-RCSID("$Id: oracle.c,v 1.84 2004/10/12 17:01:04 mast Exp $");
+RCSID("$Id: oracle.c,v 1.86 2005/08/17 09:30:04 grubba Exp $");
 
 
 /* User-changable defines: */
@@ -83,18 +83,38 @@ RCSID("$Id: oracle.c,v 1.84 2004/10/12 17:01:04 mast Exp $");
 /* #define REUSE_DEFINES */
 
 /*
- * This define crippes the Pike module to use static define buffers.
+ * This define cripples the Pike module to use static define buffers.
  * It may be required to work around bugs in some versions of the
  * oracle libraries - Hubbe
+ *
+ * Note: Some versions of Oracle fail with
+ *   "ORA-03106: fatal two-task communication protocolerror"
+ *   This seems to be due to having an invalid NLS_LANG on
+ *   the server. -Grubba 2005-08-05
  */
 
 /* For some reason this crashes if I make this larger than 8000
  * I suspect a static buffer somewhere in Oracle, but I have no proof.
  * -Hubbe
+ *
+ * One possible cause could be running out of stack in
+ * f_big_typed_query_create() due to the old struct bind_block
+ * being huge. I haven't tried raising the limit since shrinking
+ * struct bind_block.
+ * -Grubba 2005-08-04
  */
 
 /* #define STATIC_BUFFERS 8000 */
 
+/* This define causes dynamically sized data to be fetched via the polling
+ * API rather than by the callback API. Using the polling API has the
+ * advantage of allowing the OCIStmtFetch call being run in a
+ * THREADS_ALLOW() context.
+ * NOTE: Ignored if STATIC_BUFFERS above is enabled.
+ */
+
+#define POLLING_FETCH
+
 /* End user-changable defines */
 
 
@@ -160,7 +180,7 @@ DEFINE_MUTEX(oracle_serialization_mutex);
 #define tOr(X,Y) X "|" Y
 #define tArr(X) "array(" X ")"
 #define tFunc(X,Y) "function(" X ":" Y ")"
-#define tFuncV(X,Z,Y) "function(" X Z "...:" Y ")"
+#define tFuncV(X,Z,Y) "function(" X " " Z "...:" Y ")"
 #define tComma ","
 
 #define string_builder dynamic_buffer_s
@@ -406,11 +426,11 @@ struct inout
     char shortstr[32];
     OCIDate date;
     OCINumber num;
+    OCILobLocator *lob;
 #ifdef STATIC_BUFFERS
     char str[STATIC_BUFFERS];
 #endif
   } u;
-  OCILobLocator *lob;
 };
 
 static void free_inout(struct inout *i);
@@ -429,7 +449,6 @@ struct dbcon
   DEFINE_MUTEX(lock);
 
   int resultobject_busy;
-  int timeout_limit;
 };
 
 static void init_dbcon_struct(struct object *o)
@@ -442,7 +461,6 @@ static void init_dbcon_struct(struct object *o)
 #endif
   THIS_DBCON->error_handle=0;
   THIS_DBCON->context=0;
-  THIS_DBCON->timeout_limit = 30; /* default value */
   THIS_DBCON->resultobject_busy = 0;
 #ifdef LOCAL_ENV
   THIS_DBCON->env=0;
@@ -672,6 +690,12 @@ static void ora_error_handler(OCIError *err, sword rc, char *func)
 #endif
 
   OCIErrorGet(err,1,0,&errcode,msgbuf,sizeof(msgbuf),OCI_HTYPE_ERROR);
+
+#ifdef ORACLE_DEBUG
+  fprintf(stderr,"%s:code=%d:errcode=%d:%s\n",
+	  func?func:"Oracle", rc, errcode, msgbuf);
+#endif
+
   if(func)
     Pike_error("%s:code=%d:%s",func,rc,msgbuf);
   else
@@ -748,8 +772,14 @@ static sb4 output_callback(struct inout *inout,
 			   ub2 **rcodepp)
 {
 #ifdef ORACLE_DEBUG
-  fprintf(stderr,"%s(%d)\n",__FUNCTION__,inout->ftype);
+  fprintf(stderr,"%s(inout %p[%d], index %d, bufpp %p[%p], alenpp %p[%p[%d]], piecep %p[%d], indpp %p[%p], rcodepp %p[%d])\n",
+	  __FUNCTION__, inout, inout->ftype, index, bufpp, *bufpp,
+	  alenpp, *alenpp, (*alenpp)?(**alenpp):0,
+	  piecep, *piecep, indpp, *indpp, rcodepp, *rcodepp);
 #endif
+
+  CHECK_INTERPRETER_LOCK();
+
   debug_malloc_touch(bufpp);
   debug_malloc_touch(*bufpp);
   debug_malloc_touch(alenpp);
@@ -764,6 +794,11 @@ static sb4 output_callback(struct inout *inout,
   *rcodepp=&inout->rcode;
   *alenpp=&inout->xlen;
 
+#ifdef ORACLE_DEBUG
+  fprintf(stderr, "  indicator:%p (%p), rcode: %d (%p), xlen: %d (%p)\n",
+	  inout->indicator, *indpp, inout->rcode, *rcodepp, inout->xlen, *alenpp);
+#endif
+
   switch(inout->ftype)
   {
     default:
@@ -777,14 +812,26 @@ static sb4 output_callback(struct inout *inout,
     case SQLT_LNG:
       if(!STRING_BUILDER_STR(inout->output))
       {
+#ifdef ORACLE_DEBUG
+	fprintf(stderr, "New string builder.\n");
+#endif
 	init_string_builder(& inout->output,0);
       }else{
-	STRING_BUILDER_LEN(inout->output)+=inout->xlen-BLOCKSIZE;
-	inout->xlen=0;
+#ifdef ORACLE_DEBUG
+	fprintf(stderr, "Grow string builder (%d + %d).\n",
+		STRING_BUILDER_LEN(inout->output), inout->xlen);
+#endif
+	STRING_BUILDER_LEN(inout->output)+=inout->xlen;
       }
       
-      inout->xlen=BLOCKSIZE;
-      *bufpp=string_builder_allocate(& inout->output, inout->xlen, 0);
+      inout->xlen = BLOCKSIZE;
+      *bufpp = string_builder_allocate (&inout->output, inout->xlen, 0);
+      STRING_BUILDER_LEN(inout->output) -= inout->xlen;
+#ifdef ORACLE_DEBUG
+      fprintf(stderr, "Grown string builder: %d (malloced: %d).\n",
+	      STRING_BUILDER_LEN(inout->output),
+	      inout->output.malloced);
+#endif
       *piecep = OCI_NEXT_PIECE;
 #ifdef ORACLE_DEBUG
       MEMSET(*bufpp, '#', inout->xlen);
@@ -794,8 +841,8 @@ static sb4 output_callback(struct inout *inout,
 
   case SQLT_CLOB:
   case SQLT_BLOB:
-    *bufpp=inout->lob;
-    inout->xlen=sizeof(inout->lob); /* ? */
+    *bufpp=inout->u.lob;
+    inout->xlen=sizeof(inout->u.lob); /* ? */
     *piecep = OCI_ONE_PIECE;
     return OCI_CONTINUE;
 
@@ -811,7 +858,7 @@ static sb4 output_callback(struct inout *inout,
       *piecep = OCI_ONE_PIECE;
       return OCI_CONTINUE;
 
-    case SQLT_NUM:
+    case SQLT_VNU:
       *bufpp=&inout->u.num;
       inout->xlen=sizeof(inout->u.num);
       *piecep = OCI_ONE_PIECE;
@@ -828,20 +875,22 @@ static sb4 output_callback(struct inout *inout,
 
 }
 			   
-
-static sb4 define_callback(void *dbresultinfo,
+/* NOTE: May be called by OCIStmtFetch() in a THREADS_ALLOW context. */
+static sb4 define_callback(dvoid *dbresultinfo,
 			   OCIDefine *def,
 			   ub4 iter,
-			   void **bufpp,
+			   dvoid **bufpp,
 			   ub4 **alenpp,
 			   ub1 *piecep,
 			   dvoid **indpp,
 			   ub2 **rcodep)
 {
+  sb4 res;
 #ifdef ORACLE_DEBUG
   fprintf(stderr,"%s ..",__FUNCTION__);
 #endif
-  return
+
+  res =
     output_callback( &((struct dbresultinfo *)dbresultinfo)->data,
 		     iter,
 		     bufpp,
@@ -849,6 +898,12 @@ static sb4 define_callback(void *dbresultinfo,
 		     piecep,
 		     indpp,
 		     rcodep);
+
+#ifdef ORACLE_DEBUG
+  fprintf(stderr, "  ==> %d (buf: %p[%d])\n",
+	  res, *bufpp, **alenpp);
+#endif
+  return res;
 }
 
 
@@ -948,8 +1003,9 @@ static void f_fetch_fields(INT32 args)
 	ora_error_handler(dbcon->error_handle, rc, errfunc);
 
 #ifdef ORACLE_DEBUG
-      name[namelen]=0;
-      fprintf(stderr,"FIELD: name=%s length=%d type=%d\n",name,size,type);
+      /* name[namelen]=0; */
+      fprintf(stderr,"FIELD: name=%.*s length=%d type=%d\n",
+	      namelen,name,size,type);
 #endif
 
 
@@ -988,7 +1044,7 @@ static void f_fetch_fields(INT32 args)
 /* 	    OCINumberSetZero(dbcon->error_handle, &info->data.u.num); */
 	    MEMSET(&info->data.u.num, 0, data_size);
 #endif
-	    type=SQLT_NUM;
+	    type=SQLT_VNU;
 #endif
 	  }
 	  break;
@@ -1018,9 +1074,10 @@ static void f_fetch_fields(INT32 args)
 	  type_name = "blob";
 	else
 	  type_name="clob";
+	info->data.u.lob = 0;
 	if ((rc = OCIDescriptorAlloc(
 				    (dvoid *) get_oracle_environment(),
-				    (dvoid **) &info->data.lob, 
+				    (dvoid **) &info->data.u.lob, 
 				    (ub4)OCI_DTYPE_LOB, 
 				    (size_t) 0, 
 				    (dvoid **) NULL)))
@@ -1028,11 +1085,11 @@ static void f_fetch_fields(INT32 args)
 #ifdef ORACLE_DEBUG
 	    fprintf(stderr,"OCIDescriptorAlloc failed!\n");
 #endif
-	    info->data.lob = 0;
+	    info->data.u.lob = 0;
 	    info->define_handle = 0;
 	    ora_error_handler(dbcon->error_handle, rc, "OCIDescriptorAlloc");
 	  }
-	data_size=sizeof(info->data.lob); /* ? */
+	data_size=sizeof(info->data.u.lob); /* ? */
 	break;
 
 	case SQLT_RID:
@@ -1081,13 +1138,15 @@ static void f_fetch_fields(INT32 args)
 			&info->define_handle,
 			dbcon->error_handle,
 			i+1,
-			info->data.lob ? 
-			(void *)(&info->data.lob):
-			(void *)(&info->data.u),
+			/* NOTE: valuep is ignored in
+			 * OCI_DYNAMIC_FETCH mode. */
+			&info->data.u,
 #ifdef STATIC_BUFFERS
 			data_size<0? STATIC_BUFFERS :data_size,
 #else
-			data_size<0? 8000 :data_size,
+			/* But value_sz is used as the maximum piece
+			 * size if OCIDefineDynamic() is used. */
+			data_size<0? BLOCKSIZE :data_size,
 #endif
 			type,
 			& info->data.indicator,
@@ -1096,7 +1155,8 @@ static void f_fetch_fields(INT32 args)
 #ifdef STATIC_BUFFERS
 			0
 #else
-			OCI_DYNAMIC_FETCH
+			/* No need to use callbacks for fixed-length fields. */
+			data_size<0? OCI_DYNAMIC_FETCH :0
 #endif
 	);
 
@@ -1110,13 +1170,15 @@ static void f_fetch_fields(INT32 args)
       if(!IS_SUCCESS(rc))
 	ora_error_handler(dbcon->error_handle, rc, "OCIDefineByPos");
 
-#ifndef STATIC_BUFFERS
-      rc=OCIDefineDynamic(info->define_handle,
-			  dbcon->error_handle,
-			  info,
-			  define_callback);
-      if(!IS_SUCCESS(rc))
-	ora_error_handler(dbcon->error_handle, rc, "OCIDefineDynamic");
+#if !defined(STATIC_BUFFERS) && !defined(POLLING_FETCH)
+      if (data_size < 0) {
+	rc=OCIDefineDynamic(info->define_handle,
+			    dbcon->error_handle,
+			    info,
+			    define_callback);
+	if(!IS_SUCCESS(rc))
+	  ora_error_handler(dbcon->error_handle, rc, "OCIDefineDynamic");
+      }
 #endif
       debug_malloc_touch(dbcon->error_handle);
       debug_malloc_touch(dbquery->statement);
@@ -1140,9 +1202,11 @@ static void push_inout_value(struct inout *inout,
   sb4 bsize=100;
   int rslt;
   char *errfunc=0;
+  sword ret;
   
 #ifdef ORACLE_DEBUG
-  fprintf(stderr,"%s .. (type = %d, indicator = %d, len= %d)\n",__FUNCTION__,inout->ftype,inout->indicator, inout->xlen);
+  fprintf(stderr,"%s .. (type = %d, indicator = %d, len= %d)\n",
+	  __FUNCTION__,inout->ftype,inout->indicator, inout->xlen);
 #endif
 
   if(inout->indicator)
@@ -1167,6 +1231,7 @@ static void push_inout_value(struct inout *inout,
 	break;
 	
       case SQLT_NUM:
+      case SQLT_VNU:
       case SQLT_INT:
 	ref_push_object(nullint_object);
 	break;
@@ -1199,10 +1264,21 @@ static void push_inout_value(struct inout *inout,
 	break;
       }
 #endif
-      STRING_BUILDER_LEN(inout->output)+=inout->xlen-BLOCKSIZE;
+      STRING_BUILDER_LEN(inout->output) += inout->xlen;
       if(inout->ftype == SQLT_STR) 
 	STRING_BUILDER_LEN(inout->output)--;
+
+#if 0
+      for(ret=0;ret<STRING_BUILDER_LEN(inout->output);ret++)
+	fprintf(stderr,"%02x ",((unsigned char *)inout->output.s->str)[ret]);
+      fprintf(stderr,"\n");
+#endif
+
       inout->xlen=0;
+#ifdef ORACLE_DEBUG
+      fprintf(stderr, "  STRING_BUILDER_LEN(inout->output): %d\n",
+	      STRING_BUILDER_LEN(inout->output));
+#endif
       push_string(finish_string_builder(& inout->output));
       STRING_BUILDER_STR(inout->output)=0;;
       break;
@@ -1210,9 +1286,8 @@ static void push_inout_value(struct inout *inout,
     case SQLT_CLOB:
     case SQLT_BLOB:
     {
-      sword ret;
       if((ret = OCILobGetLength(dbcon->context, dbcon->error_handle,
-				inout->lob, &loblen)) != OCI_SUCCESS) {
+				inout->u.lob, &loblen)) != OCI_SUCCESS) {
 #ifdef ORACLE_DEBUG
 	fprintf(stderr,"OCILobGetLength failed.\n");
 #endif
@@ -1222,7 +1297,7 @@ static void push_inout_value(struct inout *inout,
 	if((bufp = malloc(loblen))) {
 	  if((ret = OCILobRead(dbcon->context,
 			       dbcon->error_handle,
-			       inout->lob, 
+			       inout->u.lob, 
 			       &amtp, 
 			       1, 
 			       (dvoid *) bufp,
@@ -1256,31 +1331,39 @@ static void push_inout_value(struct inout *inout,
 #if 0
       /*  Handle automatically freed when environment handle is deallocated.
 	  Not needed according according to doc.*/
-      if(inout->lob)
-	OCIDescriptorFree((dvoid *) inout->lob, (ub4) OCI_DTYPE_LOB);
+      /* WARNING: Do not enable! */
+      if(inout->u.lob)
+	OCIDescriptorFree((dvoid *) inout->u.lob, (ub4) OCI_DTYPE_LOB);
 #endif
       }
       break;
 	
     case SQLT_ODT:
     case SQLT_DAT:
+#if 0
+      for(ret=0;ret<sizeof(inout->u.date);ret++)
+	fprintf(stderr,"%02x ",((unsigned char *)&inout->u.date)[ret]);
+      fprintf(stderr,"\n");
+#endif
+
       push_object(low_clone(Date_program));
       call_c_initializers(Pike_sp[-1].u.object);
       ((struct dbdate *)STORAGE(Pike_sp[-1].u.object))->date = inout->u.date;
       break;
 
     case SQLT_NUM:
+      /* Kluge -- Convert it to a VNU. */
+      MEMMOVE(inout->u.shortstr+1,inout->u.shortstr,inout->xlen);
+      inout->u.shortstr[0]=inout->xlen;
+
+      /* FALL_THROUGH */
+    case SQLT_VNU:
     {
 #ifdef INT64
       INT64 integer;
 #else
       INT32 integer;
 #endif
-      sword ret;
-
-      /* Kluge */
-      MEMMOVE(inout->u.shortstr+1,inout->u.shortstr,inout->xlen);
-      inout->u.shortstr[0]=inout->xlen;
 
 #if 0
       for(ret=0;ret<22;ret++)
@@ -1363,7 +1446,7 @@ static void free_inout(struct inout *i)
 
 static void f_fetch_row(INT32 args)
 {
-  int i;
+  int i = 0;
   sword rc;
   struct dbresult *dbresult = THIS_RESULT;
   struct dbquery *dbquery = THIS_RESULT_QUERY;
@@ -1381,25 +1464,148 @@ static void f_fetch_row(INT32 args)
     pop_stack();
   }
 
-  THREADS_ALLOW();
-#ifdef ORACLE_DEBUG
-  fprintf(stderr,"OCIStmtFetch\n");
+  do {
+#if defined(STATIC_BUFFERS) || defined(POLLING_FETCH)
+    /* NOTE: output_callback() is currently not safe to run
+     *       in a THREADS_ALLOW() context.
+     */
+    THREADS_ALLOW();
 #endif
-  rc=OCIStmtFetch(dbquery->statement,
-		  dbcon->error_handle,
-		  1,
-		  OCI_FETCH_NEXT,
-		  OCI_DEFAULT);
 #ifdef ORACLE_DEBUG
-  fprintf(stderr,"OCIStmtFetch done\n");
+    fprintf(stderr,"OCIStmtFetch\n");
+    {
+      static text msgbuf[512];
+      ub4 errcode;
+      OCIErrorGet(dbcon->error_handle,1,0,&errcode,
+		  msgbuf,sizeof(msgbuf),OCI_HTYPE_ERROR);
+      fprintf(stderr, "  Before: errcode=%d:%s\n", errcode, msgbuf);
+    }
+#endif
+    rc=OCIStmtFetch(dbquery->statement,
+		    dbcon->error_handle,
+		    1,
+		    OCI_FETCH_NEXT,
+		    OCI_DEFAULT);
+#ifdef ORACLE_DEBUG
+    fprintf(stderr,"OCIStmtFetch done rc=%d\n", rc);
+#endif
+#if defined(STATIC_BUFFERS) || defined(POLLING_FETCH)
+    THREADS_DISALLOW();
 #endif
-  THREADS_DISALLOW();
 
-  if(rc==OCI_NO_DATA)
-  {
-    push_int(0);
-    return;
-  }
+    if(rc==OCI_NO_DATA)
+    {
+      push_int(0);
+      return;
+    }
+#ifdef POLLING_FETCH
+    if (rc == OCI_NEED_DATA) {
+      OCIDefine *define;
+      ub4 htype;
+      ub1 direction;
+      ub4 iter;
+      ub4 index;
+      ub1 piece;
+      ub4 ret = OCIStmtGetPieceInfo(dbquery->statement,
+				    dbcon->error_handle,
+				    (void **)&define,
+				    &htype,
+				    &direction,
+				    &iter,
+				    &index,
+				    &piece);
+      struct dbresultinfo *info = NULL;
+      struct inout *inout;
+      char *buf;
+
+      if (!IS_SUCCESS(ret))
+	ora_error_handler(dbcon->error_handle, ret, "OCIStmtGetPieceInfo");
+#ifdef ORACLE_DEBUG
+      fprintf(stderr, "OCIStmtGetPieceInfo ==>\n"
+	      "  define: %p\n"
+	      "  htype: %d\n"
+	      "  direction: %d\n"
+	      "  iter: %d\n"
+	      "  index: %d\n"
+	      "  piece: %d\n",
+	      define, htype, direction, iter, index, piece);
+#endif
+      if (htype != OCI_HTYPE_DEFINE)
+	break;	/* Not supported. */
+      /* NOTE: Columns come in order, so there's no need to
+       *       rescan the first columns in every pass.
+       */
+      for (; i < dbquery->field_info->size; i++) {
+	if (dbquery->field_info->item[i].type == T_OBJECT) {
+	  struct object *o = dbquery->field_info->item[i].u.object;
+	  if (o->prog == dbresultinfo_program) {
+	    struct dbresultinfo *in = (struct dbresultinfo *)STORAGE(o);
+	    if (in->define_handle == define) {
+	      info = in;
+#ifdef ORACLE_DEBUG
+	      fprintf(stderr, "Found info %p for define %p (i:%d)\n",
+		      in, define, i);
+#endif
+	      break;
+	    }
+	  }
+	}
+      }
+      if (!info) {
+	/* Not found! */
+#ifdef ORACLE_DEBUG
+	fprintf(stderr, "Failed to find info for define %p\n",
+		define);
+#endif
+	break;
+      }
+      inout = &info->data;
+      if ((inout->ftype != SQLT_LNG) && (inout->ftype != SQLT_LBI)) {
+	/* Unsupported */
+#ifdef ORACLE_DEBUG
+	fprintf(stderr, "Piecewise access for ftype %d not supported.\n",
+		inout->ftype);
+#endif
+	break;
+      }
+      if(!STRING_BUILDER_STR(inout->output))
+      {
+#ifdef ORACLE_DEBUG
+	fprintf(stderr, "New string builder.\n");
+#endif
+	init_string_builder(& inout->output,0);
+      }else{
+#ifdef ORACLE_DEBUG
+	fprintf(stderr, "Grow string builder (%d + %d).\n",
+		STRING_BUILDER_LEN(inout->output), inout->xlen);
+#endif
+	STRING_BUILDER_LEN(inout->output)+=inout->xlen;
+      }
+      
+      inout->xlen = BLOCKSIZE;
+      buf = string_builder_allocate (&inout->output, inout->xlen, 0);
+      STRING_BUILDER_LEN(inout->output) -= inout->xlen;
+#ifdef ORACLE_DEBUG
+      fprintf(stderr, "Grown string builder: %d (malloced: %d).\n",
+	      STRING_BUILDER_LEN(inout->output),
+	      inout->output.malloced);
+#endif
+      /* piece = OCI_NEXT_PIECE; */
+#ifdef ORACLE_DEBUG
+      MEMSET(buf, '#', inout->xlen);
+      buf[inout->xlen-1]=0;
+#endif
+      ret = OCIStmtSetPieceInfo(define, htype, dbcon->error_handle,
+				buf, &inout->xlen, piece,
+				&inout->indicator,
+				&inout->rcode);
+      if(!IS_SUCCESS(ret))
+	ora_error_handler(dbcon->error_handle, rc, "OCIStmtSetPieceInfo");
+    }
+#else /* !POLLING_FETCH */
+    break;
+#endif /* POLLING_FETCH */
+  } while (rc == OCI_NEED_DATA);
 
   if(!IS_SUCCESS(rc))
     ora_error_handler(dbcon->error_handle, rc, "OCIStmtFetch");
@@ -1524,27 +1730,6 @@ static void f_oracle_create(INT32 args)
   return;
 }
 
-static void f_dbcon_timeout_limit(INT32 args)
-{
-  struct dbcon *dbcon=THIS_DBCON;
-  /* No arguments: get timeout. One argument: set timeout. */
-#ifdef ORACLE_DEBUG
-  fprintf(stderr, "%s, dbcon=%p, args=%d\n", __FUNCTION__, dbcon, args);
-#endif
-  if (args)
-  {
-    INT_TYPE new_timeout;
-    get_all_args("Oracle->timeout", args, "%i", &new_timeout);
-    if (new_timeout < 0)
-      Pike_error("Negative timeout specified.\n");
-    if (new_timeout > 3600) /* max one hour */
-      new_timeout = 3600;
-    dbcon->timeout_limit = new_timeout;
-  }
-  pop_n_elems(args);
-  push_int(dbcon->timeout_limit);
-}
-
 static void f_compile_query_create(INT32 args)
 {
   int rc;
@@ -1568,21 +1753,8 @@ static void f_compile_query_create(INT32 args)
 
   if (dbcon->resultobject_busy)
   {
-    int t, timeout_limit = threads_disabled ? 0 : dbcon->timeout_limit;
-
-    for(t = 0; t < timeout_limit && dbcon->resultobject_busy; ++t)
-    {
-      THREADS_ALLOW();
-#ifdef __NT__
-      Sleep(1000);
-#else
-      sleep(1);
-#endif
-      THREADS_DISALLOW();
-    }
-    if (dbcon->resultobject_busy)
-      Pike_error("Oracle connection busy; previous result object "
-		 "still active.\n");
+    Pike_error("Oracle connection busy; previous result object "
+	       "still active.\n");
   }
 
   THREADS_ALLOW();
@@ -1650,15 +1822,12 @@ struct bind
   struct inout data;
 };
 
-#define MAX_NUMBER_OF_BINDINGS 256
-
 struct bind_block
 {
-  struct bind bind[MAX_NUMBER_OF_BINDINGS];
+  struct bind *bind;
   int bindnum;
 };
 
-
 static sb4 input_callback(void *vbind_struct,
 			   OCIBind *bindp,
 			   ub4 iter,
@@ -1721,6 +1890,10 @@ static void free_bind_block(struct bind_block *bind)
     free_inout(& bind->bind[bind->bindnum].data);
     bind->bindnum--;
   }
+  if (bind->bind) {
+    free(bind->bind);
+    bind->bind = NULL;
+  }
 }
 
 /*
@@ -1739,11 +1912,8 @@ static void f_big_typed_query_create(INT32 args)
   int i,num;
   struct object *new_parent=0;
   struct bind_block bind;
-
   extern int d_flag;
 
-  bind.bindnum=-1;
-
 #ifdef ORACLE_DEBUG
   fprintf(stderr,"%s\n",__FUNCTION__);
 #endif
@@ -1777,8 +1947,14 @@ static void f_big_typed_query_create(INT32 args)
     case 0: break;
   }
 
-  if(bnds && m_sizeof(bnds) > MAX_NUMBER_OF_BINDINGS)
-    Pike_error("Too many variables.\n");
+  bind.bindnum=-1;    
+  bind.bind = NULL;
+
+  SET_ONERROR(err, free_bind_block, &bind);
+
+  if (bnds && m_sizeof(bnds)) {
+    bind.bind = xalloc(sizeof(struct bind) * m_sizeof(bnds));
+  }
 
   destruct_objects_to_destruct();
 
@@ -1810,13 +1986,15 @@ static void f_big_typed_query_create(INT32 args)
   }
 #endif
 
-  SET_ONERROR(err, free_bind_block, &bind);
-
-  if(bnds != NULL)
+  if(bnds && m_sizeof(bnds))
   {
     INT32 e;
     struct keypair *k;
     struct mapping_data *md = bnds->data;
+#ifdef ORACLE_DEBUG
+    fprintf(stderr, "Binding %d variables...\n",
+	    m_sizeof(bnds));
+#endif
     NEW_MAPPING_LOOP(md)
       {
 	struct svalue *value=&k->val;
@@ -2049,8 +2227,7 @@ static void f_big_typed_query_create(INT32 args)
     }
   }
 
-  free_bind_block(&bind);
-  UNSET_ONERROR(err);
+  CALL_AND_UNSET_ONERROR(err);
 
 #ifdef _REENTRANT
   if(d_flag)
@@ -2192,6 +2369,8 @@ static void dbnull_not(INT32 args)
 PIKE_MODULE_INIT
 {
   long offset=0;
+  sword ret;
+
 #ifdef ORACLE_HOME
   if(getenv("ORACLE_HOME")==NULL)
     putenv("ORACLE_HOME="ORACLE_HOME);
@@ -2205,13 +2384,23 @@ PIKE_MODULE_INIT
   fprintf(stderr,"%s\n",__FUNCTION__);
 #endif
 
-  
-  if(OCIInitialize(
-    OCI_OBJECT | ORACLE_INIT_FLAGS,
-    0, ocimalloc, ocirealloc, ocifree) != OCI_SUCCESS)
+  if ((ret = 
+#if 0
+       /* Oracle 9 */
+       OCIEnvCreate( &oracle_environment,
+		     OCI_OBJECT | ORACLE_INIT_FLAGS,
+		     NULL,
+		     ocimalloc, ocirealloc, ocifree,
+		     0, NULL)
+#else
+       /* Oracle 8 */
+       OCIInitialize( OCI_OBJECT | ORACLE_INIT_FLAGS,
+		      0, ocimalloc, ocirealloc, ocifree)
+#endif
+       ) != OCI_SUCCESS)
   {
 #ifdef ORACLE_DEBUG
-    fprintf(stderr,"OCIInitialize failed\n");
+    fprintf(stderr,"OCIInitialize failed: %d\n", ret);
 #endif
     return;
   }
@@ -2281,9 +2470,6 @@ PIKE_MODULE_INIT
       MY_END_CLASS(compile_query);
     }
 
-    ADD_FUNCTION("timeout_limit", f_dbcon_timeout_limit,
-		 tFunc(tOr(tInt,tVoid), tInt), ID_PUBLIC);
-
     ADD_FUNCTION("create", f_oracle_create,tFunc(tOr(tStr,tVoid) tComma tOr(tStr,tVoid) tComma tOr(tStr,tVoid) tComma tOr(tStr,tVoid),tVoid), ID_PUBLIC);
     
     MY_END_CLASS(oracle);
