@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: pg_types.h,v 1.6 2002/10/11 01:39:48 nilsson Exp $
+|| $Id: pg_types.h,v 1.7 2003/12/15 22:28:24 nilsson Exp $
 */
 
 #ifndef _PG_TYPES_H_
@@ -14,6 +14,7 @@
 struct postgres_result_object_data {
   PGresult * result;
   int cursor;
+  struct pgres_object_data* pgod;
 };
 
 struct pgres_object_data {
@@ -21,6 +22,9 @@ struct pgres_object_data {
 	struct pike_string *last_error;
 	PGresult * last_result;
 	struct svalue * notify_callback;
+	int dofetch;
+	int docommit;
+	int lastcommit;
 #ifdef PQ_THREADSAFE
         PIKE_MUTEX_T mutex;
 #endif
@@ -29,4 +33,8 @@ struct pgres_object_data {
 /* The header name could be deceiving, but who cares? */
 extern struct program *postgres_program, *pgresult_program;
 
+#define FETCHSIZESTR	"64"
+#define CURSORNAME	"_pikecursor"
+#define FETCHCMD	"FETCH " FETCHSIZESTR " IN " CURSORNAME
+
 #endif
