@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: postgres.c,v 1.39 2005/07/16 17:21:14 nilsson Exp $
+|| $Id: postgres.c,v 1.42 2006/06/06 03:26:00 adam Exp $
 */
 
 /*
@@ -72,7 +72,7 @@ static void pgdebug (char * a, ...) {}
 
 struct program * postgres_program;
 
-RCSID("$Id: postgres.c,v 1.39 2005/07/16 17:21:14 nilsson Exp $");
+RCSID("$Id: postgres.c,v 1.42 2006/06/06 03:26:00 adam Exp $");
 
 static void set_error (char * newerror)
 {
@@ -728,6 +728,39 @@ static void f_host_info (INT32 args)
 	Pike_error ("Bad connection.\n");
 }
 
+/*! @decl string _quote(string s)
+ *!
+ *! Escape a string to prevent SQL injection, using the current connection's
+ *! character encoding settings.
+ */
+#if defined(HAVE_PQESCAPESTRINGCONN) || defined(HAVE_PQESCAPESTRING)
+static void f_quote(INT32 args)
+{
+	int err;
+	int len;
+	struct pike_string *ret;
+	struct pike_string *s;
+	char *err_msg;
+
+	get_all_args("Postgres->_quote", args, "%S", &s);
+
+	ret = begin_shared_string(s->len * 2 + 1);
+#ifdef HAVE_PQESCAPESTRINGCONN
+	len = PQescapeStringConn(THIS->dblink, ret->str, s->str, s->len, &err);
+	if (err != 0) {
+		err_msg = PQerrorMessage(THIS->dblink);
+		set_error(err_msg);
+		Pike_error(err_msg);
+	}
+#else
+	len = PQescapeString(ret->str, s->str, s->len);
+#endif
+	pop_n_elems(args);
+	push_string(end_and_resize_shared_string(ret, len));
+}
+#endif /* HAVE_PQESCAPESTRINGCONN || HAVE_PQESCAPESTRING */
+
+
 /*! @endclass
  *!
  *! @endmodule
@@ -762,6 +795,11 @@ PIKE_MODULE_INIT
   ADD_FUNCTION("host_info",f_host_info,tFunc(tVoid,tStr),
 			OPT_EXTERNAL_DEPEND|OPT_RETURN);
 
+  /* function(string:string) */
+#if defined(HAVE_PQESCAPESTRINGCONN) || defined(HAVE_PQESCAPESTRING)
+  ADD_FUNCTION("_quote", f_quote, tFunc(tStr,tStr), 0);
+#endif
+
 	/* postgres-specific functions */
 	/* function(void:void) */
   ADD_FUNCTION("reset",f_reset,tFunc(tVoid,tVoid),OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
