@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: image_jpeg.c,v 1.65 2004/04/15 00:10:17 nilsson Exp $
+|| $Id: image_jpeg.c,v 1.66 2004/07/01 11:36:03 nilsson Exp $
 */
 
 #include "global.h"
@@ -59,7 +59,7 @@
 #ifdef HAVE_STDLIB_H
 #undef HAVE_STDLIB_H
 #endif
-RCSID("$Id: image_jpeg.c,v 1.65 2004/04/15 00:10:17 nilsson Exp $");
+RCSID("$Id: image_jpeg.c,v 1.66 2004/07/01 11:36:03 nilsson Exp $");
 
 /* jpeglib defines EXTERN for some reason.
  * This is not good, since it confuses compilation.h.
@@ -166,7 +166,7 @@ struct my_decompress_struct
       struct my_marker *next;
       INT32 id;
       INT32 len;
-      char data[1];
+      unsigned char data[1];
    } *first_marker;
 };
 
@@ -448,7 +448,7 @@ static int parameter_marker(struct svalue *map,struct pike_string *what,
 	    Pike_error("Image.JPEG.encode: illegal value of option "
 		       "marker; expected mapping(int:8 bit string)\n");
 	 jpeg_write_marker(cinfo, k->ind.u.integer, 
-			   k->val.u.string->str, 
+			   (const unsigned char *)k->val.u.string->str,
 			   k->val.u.string->len); 
       }
 
@@ -468,7 +468,7 @@ static int parameter_comment(struct svalue *map,struct pike_string *what,
 		 " expected 8 bit string\n");
 
    jpeg_write_marker(cinfo, JPEG_COM, 
-		     v->u.string->str,
+		     (const unsigned char *)v->u.string->str,
 		     v->u.string->len); 
 
    return 1;
@@ -871,7 +871,7 @@ static void image_jpeg_encode(INT32 args)
 	      p==JDCT_FASTEST))
 	 cinfo.dct_method=p;
       
-      if (parameter_int(sp+1-args,param_progressive,&p))
+      if (parameter_int(sp+1-args,param_progressive,&p) && p)
 	 jpeg_simple_progression(&cinfo);
 
       parameter_qt(sp+1-args,param_quant_tables,&cinfo);
