@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: xml.c,v 1.57 2002/12/17 18:06:34 nilsson Exp $
+|| $Id: xml.c,v 1.71 2005/05/18 12:36:54 mast Exp $
 */
 
 #include "global.h"
@@ -22,7 +22,6 @@
 #include "module_support.h"
 #include "operators.h"
 #include "pike_error.h"
-#include "opcodes.h"
 #include "block_alloc.h"
 #include "bignum.h"
 
@@ -65,9 +64,14 @@ struct xmlobj
   struct mapping *entities;
   struct mapping *attributes;
   struct mapping *is_cdata;
-  int rxml_mode;
+  int flags;
 };
 
+/* Flag bits. */
+#define ALLOW_RXML_ENTITIES 0x01
+#define COMPAT_ALLOW_7_2_ERRORS 0x02
+#define COMPAT_ALLOW_7_6_ERRORS 0x04
+
 
 static struct svalue location_string_svalue;
 
@@ -649,12 +653,21 @@ ISWRAP(isHexChar)
 
 #ifdef VERBOSE_XMLDEBUG
 #define IF_XMLDEBUG(X) X
+#ifndef PIKE_DEBUG
+#define PIKE_DEBUG
+#endif
 #else
 #define IF_XMLDEBUG(X)
 #endif
 
+#ifdef PIKE_DEBUG
+#define IF_PIKEDEBUG(X) X
+#else
+#define IF_PIKEDEBUG(X)
+#endif
+
 #define PEEK(X) \
-( IF_XMLDEBUG(((X) > data->input.len)? \
+( IF_PIKEDEBUG(((X) > data->input.len)? \
     (Pike_fatal("PEEK out of bounds (%ld/%ld)\n", \
            PTRDIFF_T_TO_LONG(X), \
            PTRDIFF_T_TO_LONG(data->input.len)), 0):) \
@@ -686,6 +699,16 @@ ISWRAP(isHexChar)
                      data->input.pos, data->input.to_free));		\
 } while(0)
 
+#ifdef PIKE_DEBUG
+#define CHECK_INPUT(INPUT) do {					\
+    if ((INPUT).len < 0) {					\
+      Pike_fatal("Negative input length: %"PRINTPTRDIFFT"d\n", (INPUT).len); \
+    }								\
+  } while(0)
+#else /* !PIKE_DEBUG */
+#define CHECK_INPUT(INPUT)
+#endif /* PIKE_DEBUG */
+
 static inline int xmlread(int z,struct xmldata *data, int line)
 {
   int popped=0;
@@ -723,10 +746,12 @@ static void simple_read_system_literal(struct xmldata *);
 static void simple_read_pubid_literal(struct xmldata *);
 static int low_parse_xml(struct xmldata *data,
 			 struct pike_string *end,
-			 int toplevel);
-static void xmlerror(char *desc, struct xmldata *data);
+			 int *doc_seq_pos);
+static void xmlerror(char *desc, struct xmldata *data,
+		     struct pike_string *tag_name);
 
-#define XMLERROR(desc) xmlerror(desc,data)
+#define XMLERROR(desc) do {xmlerror(desc,data,NULL); READ (1);} while (0)
+#define XMLERROR_STAY(desc) xmlerror(desc,data,NULL)
 
 #define VOIDIFY(X) do { struct svalue *save_sp=sp; 	\
        X;						\
@@ -734,10 +759,10 @@ static void xmlerror(char *desc, struct xmldata *data);
     }while(0);
 
 #define SKIPSPACE() \
-  do { while(SMEG(), isSpace(PEEK(0))) READ(1); }while(0)
+  do { while((!XMLEOF()) && (SMEG(), isSpace(PEEK(0)))) READ(1); }while(0)
 
 #define SKIPSPACE_NO_SMEG() \
-  do { while(isSpace(PEEK(0))) READ(1); }while(0)
+  do { while((!XMLEOF()) && isSpace(PEEK(0))) READ(1); }while(0)
 
 #define SKIPTO(X) \
   do { while( (!XMLEOF()) && (PEEK(0) != (X))) READ(1); }while(0)
@@ -764,7 +789,7 @@ static int gobble(struct xmldata *data, char *s)
 	  POKE(X, PEEK(0));				\
 	  READ(1);					\
 	}else{						\
-	  XMLERROR("Name expected");		\
+	  XMLERROR("Name expected");			\
 	}						\
 	while(!XMLEOF() && isNameChar(PEEK(0)))		\
 	{						\
@@ -782,7 +807,7 @@ static int gobble(struct xmldata *data, char *s)
 	  POKE(X, PEEK(0));				\
 	  READ(1);					\
 	}else{						\
-	  XMLERROR("Name expected");		\
+	  XMLERROR("Name expected");			\
 	}						\
 	while(!XMLEOF() && isNameChar(PEEK(0)))		\
 	{						\
@@ -946,6 +971,7 @@ static int gobble(struct xmldata *data, char *s)
 	  my_tmp.input.len=s->len;					 \
 	  my_tmp.input.pos=0;						 \
 	  my_tmp.input.next=0;						 \
+	  CHECK_INPUT(my_tmp.input);					 \
 	  add_ref(my_tmp.input.callbackinfo=callbackinfo);		 \
 	  PARSE_RECURSIVELY;						 \
 	  if(THIS->entities)						 \
@@ -966,7 +992,7 @@ static int gobble(struct xmldata *data, char *s)
     READ_CHAR_REF(X);							    \
   }else{                                                                    \
     int found_period = 0;                                                   \
-    if (THIS->rxml_mode) {                                                  \
+    if (THIS->flags & ALLOW_RXML_ENTITIES) {				    \
       found_period = SIMPLE_READNAME_PERIOD();                              \
     } else {                                                                \
       SIMPLE_READNAME();                                                    \
@@ -1123,15 +1149,19 @@ static void sys(struct xmldata *data)
 
 #define SYS() sys(data)
 
-static void xmlerror(char *desc, struct xmldata *data)
+static void xmlerror(char *desc, struct xmldata *data,
+		     struct pike_string *tag_name)
 {
   push_constant_text("error");
-  push_int(0); /* no name */
+  if (tag_name) {
+    ref_push_string(tag_name);	/* Name of tag that triggered the error. */
+  } else {
+    push_int(0); /* no name */
+  }
   push_int(0); /* no attributes */
   push_text(desc);
   low_sys(data);
   pop_stack();
-  READ(1);
 }
 
 static int read_smeg_pereference(struct xmldata *data)
@@ -1212,6 +1242,7 @@ static int read_smeg_pereference(struct xmldata *data)
 	  data->input.pos=0;
 	  data->input.datap=MKPCHARP_STR(s);
 	  data->input.len=(s)->len;
+	  CHECK_INPUT(data->input);
 	  free_mapping(data->input.callbackinfo);
 	  data->input.callbackinfo=callbackinfo;
 	  copy_shared_string(data->input.to_free,s);
@@ -1491,16 +1522,26 @@ static void parse_optional_xmldecl(struct xmldata *data)
 {
   if(GOBBLE("<?xml"))
   {
+    struct mapping *m;
+
     push_constant_text("<?xml");
     push_int(0);
-    push_mapping(allocate_mapping(10)); /* Attributes */
+    push_mapping(m = allocate_mapping(10)); /* Attributes */
     
     SIMPLE_READ_ATTRIBUTES(0);
     
     if(PEEK(0) != '?' && PEEK(1)!='>')
-      XMLERROR("Missing ?> at end of <?xml.");
-    READ(2);
-	      
+      XMLERROR("Missing '?>' at end of XML header.");
+    else
+      READ(2);
+
+    if (!(THIS->flags & COMPAT_ALLOW_7_6_ERRORS)) {
+      struct pike_string *str_version;
+      MAKE_CONST_STRING (str_version, "version");
+      if (!low_mapping_string_lookup (m, str_version))
+	XMLERROR ("Required version attribute missing in XML header.");
+    }
+
     push_int(0); /* No data */
     SYS();
   }
@@ -1561,7 +1602,7 @@ void read_choice_seq_or_name(struct xmldata *data, int maybe_pcdata)
 	switch(PEEK(0))
 	{
 	  case 0:
-	    XMLERROR("End of xml while readin ELEMENT declaration.");
+	    XMLERROR("End of xml while reading ELEMENT declaration.");
 	    
 	  default:
 	    XMLERROR("Expected | or ,");
@@ -2285,7 +2326,7 @@ static int really_low_parse_dtd(struct xmldata *data)
 
 static int low_parse_dtd(struct xmldata *data)
 {
-  int done=0;
+  int done;
   struct svalue *save_sp=sp;
 
   done=really_low_parse_dtd(data);
@@ -2304,13 +2345,26 @@ static int low_parse_dtd(struct xmldata *data)
   return done;
 }  
 
+#define DOC_BEGINNING 0
+#define DOC_GOT_DOCTYPE 1
+#define DOC_IN_ROOT_ELEM 2
+#define DOC_AFTER_ROOT_ELEM 3
+
 static struct pike_string *very_low_parse_xml(struct xmldata *data,
 					      struct pike_string *end,
-					      int toplevel,
 					      struct string_builder *text,
-					      int keepspaces)
+					      int keepspaces,
+					      int *doc_seq_pos)
 {
-  int done=0;
+  int done=0, toplevel = *doc_seq_pos != DOC_IN_ROOT_ELEM;
+
+#define CHECK_TOPLEVEL_EPILOG(ERRMSG) do {				\
+    if (*doc_seq_pos == DOC_AFTER_ROOT_ELEM &&				\
+	!(THIS->flags & COMPAT_ALLOW_7_2_ERRORS)) {			\
+      XMLERROR_STAY (ERRMSG);						\
+    }									\
+  } while (0)
+
   while(!done && !XMLEOF())
   {
     switch(PEEK(0))
@@ -2319,10 +2373,7 @@ static struct pike_string *very_low_parse_xml(struct xmldata *data,
 	if(toplevel)
 	{
 	  if(!isSpace(PEEK(0)))
-	  {
 	    XMLERROR("All data must be inside tags");
-	    READ(1);
-	  }
 	  SKIPSPACE();
 	  break;
 	}
@@ -2351,7 +2402,7 @@ static struct pike_string *very_low_parse_xml(struct xmldata *data,
 	continue;
 
       case '&':
-	READ_REFERENCE(0,*text,very_low_parse_xml(&my_tmp, end, toplevel, text, 1));
+	READ_REFERENCE(0,*text,very_low_parse_xml(&my_tmp, NULL, text, 1, doc_seq_pos));
 	continue;
 
       case '<':
@@ -2360,11 +2411,14 @@ static struct pike_string *very_low_parse_xml(struct xmldata *data,
 	switch(PEEK(1))
 	{
 	  case '?': /* Ends with ?> */
-	    if(PEEK(2)=='x' &&
-	       PEEK(3)=='m' &&
-	       PEEK(4)=='l' &&
-	       isSpace(PEEK(5)))
-	    {
+	    if ((THIS->flags & COMPAT_ALLOW_7_6_ERRORS) &&
+		PEEK(2)=='x' &&
+		PEEK(3)=='m' &&
+		PEEK(4)=='l' &&
+		isSpace(PEEK(5))) {
+	      /* Since parse_optional_xmldecl is used to parse the
+	       * initial xml header, I can't see why this case
+	       * exists. /mast */
 	      push_constant_text("<?xml");
 	      READ(6);
 	      push_int(0);
@@ -2436,6 +2490,14 @@ static struct pike_string *very_low_parse_xml(struct xmldata *data,
 		   PEEK(8)=='[')
 		{
 		  READ(9);
+
+		  if (!(THIS->flags & COMPAT_ALLOW_7_6_ERRORS)) {
+		    if (toplevel)
+		      XMLERROR_STAY ("All data must be inside tags");
+		  }
+		  else
+		    CHECK_TOPLEVEL_EPILOG ("All data must be inside tags");
+
 		  push_constant_text("<![CDATA[");
 		  push_int(0);
 		  push_int(0);
@@ -2479,14 +2541,33 @@ static struct pike_string *very_low_parse_xml(struct xmldata *data,
 		  XMLERROR("Expected 'DOCTYPE', got something else.");
 		}else{
 		  READ(9);
+
+		  if (!(THIS->flags & COMPAT_ALLOW_7_6_ERRORS)) {
+		    switch (*doc_seq_pos) {
+		      case DOC_GOT_DOCTYPE:
+			XMLERROR_STAY (
+			  "Multiple DOCTYPE declarations are not allowed.");
+			break;
+		      case DOC_IN_ROOT_ELEM:
+		      case DOC_AFTER_ROOT_ELEM:
+			XMLERROR_STAY (
+			  "DOCTYPE must occur before the root element.");
+			break;
+		    }
+		  }
+		  else
+		    CHECK_TOPLEVEL_EPILOG (
+		      "DOCTYPE must occur before the root element.");
+
 		  SKIPSPACE();
 		  push_constant_text("<!DOCTYPE");
 		  SIMPLE_READNAME(); /* NAME */
 		  SKIPSPACE();
-		  switch(PEEK(0))
+		  switch(SAFE_PEEK(0))
 		  {
 		    case 'P':
-		      if(PEEK(1)=='U' &&
+		      if((data->input.len > 5) &&
+			 PEEK(1)=='U' &&
 			 PEEK(2)=='B' &&
 			 PEEK(3)=='L' &&
 			 PEEK(4)=='I' &&
@@ -2505,7 +2586,8 @@ static struct pike_string *very_low_parse_xml(struct xmldata *data,
 		      break;
 
 		    case 'S':
-		      if(PEEK(1)=='Y' &&
+		      if((data->input.len > 5) &&
+			 PEEK(1)=='Y' &&
 			 PEEK(2)=='S' &&
 			 PEEK(3)=='T' &&
 			 PEEK(4)=='E' &&
@@ -2525,7 +2607,7 @@ static struct pike_string *very_low_parse_xml(struct xmldata *data,
 		      f_aggregate_mapping(0);
 		  }
 
-		  if(PEEK(0)=='[')
+		  if(SAFE_PEEK(0)=='[')
 		  {
 		    READ(1);
 		    low_parse_dtd(data);
@@ -2533,17 +2615,20 @@ static struct pike_string *very_low_parse_xml(struct xmldata *data,
 		    fprintf(stderr,"FOO: %c%c%c%c\n",SAFE_PEEK(0),
 			    SAFE_PEEK(1),SAFE_PEEK(2),SAFE_PEEK(3));
 #endif
-		    if(PEEK(0) != ']')
+		    if(SAFE_PEEK(0) != ']')
 		      XMLERROR("Missing ] in DOCTYPE tag.");
 		    READ(1);
 		    SKIPSPACE();
 		  }else{
 		    push_int(0);
 		  }
-		  if(PEEK(0)!='>')
+		  if(SAFE_PEEK(0)!='>')
 		    XMLERROR("Missing '>' in DOCTYPE tag.");
 		  READ(1);
+
 		  SYS();
+		  if (*doc_seq_pos < DOC_GOT_DOCTYPE)
+		    *doc_seq_pos = DOC_GOT_DOCTYPE;
 		}
 		break;
 	    }
@@ -2554,12 +2639,12 @@ static struct pike_string *very_low_parse_xml(struct xmldata *data,
 	    SIMPLE_READNAME();
 	    SKIPSPACE();
 	    if(PEEK(0)!='>')
-	      XMLERROR("Missing > in end tag.");
+	      xmlerror("Missing > in end tag.", data, sp[-1].u.string);
 	    else
 	      READ(1);
 	    if(end!=sp[-1].u.string)
 	    {
-	      XMLERROR("Unmatched end tag.");
+	      xmlerror("Unmatched end tag.", data, sp[-1].u.string);
 	    }else{
 	      end=0;
 	    }
@@ -2569,6 +2654,9 @@ static struct pike_string *very_low_parse_xml(struct xmldata *data,
 
 	  default:
 	    /* 'Normal' tag (we hope) */
+	    CHECK_TOPLEVEL_EPILOG (
+	      "There can not be more than one element on the top level.");
+
 	    push_constant_text(">"); 
 
 	    READ(1);
@@ -2601,10 +2689,17 @@ static struct pike_string *very_low_parse_xml(struct xmldata *data,
 	      SIMPLE_READ_ATTRIBUTES(m);
 	    }
 
+	    /* At this point the stack contains the following:
+	     *
+	     * sp[-3]: ">".
+	     * sp[-2]: tag_name.
+	     * sp[-1]: attributes.
+	     */
+
 	    switch(PEEK(0))
 	    {
 	      default:
-		XMLERROR("Failed to find end of tag.");
+		xmlerror("Failed to find end of tag.", data, sp[-2].u.string);
 		pop_n_elems(3);
 		break;
 
@@ -2623,32 +2718,31 @@ static struct pike_string *very_low_parse_xml(struct xmldata *data,
 		  pop_n_elems(sp-save_sp);
 		}
 
-		if(low_parse_xml(data, sp[-2].u.string,0))
-		  XMLERROR("Unmatched tag.");
+		if (toplevel) *doc_seq_pos = DOC_IN_ROOT_ELEM;
+		if(low_parse_xml(data, sp[-2].u.string, doc_seq_pos)) {
+		  /* NOTE: low_parse_xml() pushes an element on the stack. */
+		  xmlerror("Unmatched tag.", data, sp[-3].u.string);
+		}
 		SYS();
+		if (toplevel) *doc_seq_pos = DOC_AFTER_ROOT_ELEM;
 		break;
 
 	      case '/':
 		READ(1);
 		if(PEEK(0)!='>')
-		  XMLERROR("Missing '>' in empty tag.");
-		READ(1);
+		  xmlerror("Missing '>' in empty tag.", data,
+			   sp[-2].u.string);
+		else
+		  READ(1);
 		/* Self-contained tag */
 		free_string(sp[-3].u.string);
 		sp[-3].u.string=make_shared_string("<>");
 		push_int(0); /* No data */
 		SYS();
+		if (toplevel) *doc_seq_pos = DOC_AFTER_ROOT_ELEM;
 		break;
 		
 	    }
-
-	    if(toplevel)
-	    {
-	      done=1;
-	      SKIPSPACE();
-	      if(!XMLEOF())
-		XMLERROR("All data must be in one top-level tag.\n");
-	    }
 	}
     }
   }
@@ -2657,12 +2751,11 @@ static struct pike_string *very_low_parse_xml(struct xmldata *data,
 
 static int low_parse_xml(struct xmldata *data,
 			 struct pike_string *end,
-			 int toplevel)
+			 int *doc_seq_pos)
 {
   struct svalue *save_sp=sp;
   BEGIN_STRING(text);
-  parse_optional_xmldecl(data);
-  end=very_low_parse_xml(data,end,toplevel,&text,0);
+  end=very_low_parse_xml(data,end,&text, 0, doc_seq_pos);
   INTERMISSION(text);
   END_STRING(text);
   pop_stack();
@@ -2678,7 +2771,11 @@ static int low_parse_xml(struct xmldata *data,
 /*! @module spider
  */
 
-/*! @decl array parse_xml(string xml, function cb)
+/*! @class XML
+ *! @appears Parser.XML.Simple
+ */
+
+/*! @decl array parse(string xml, function cb)
  */
 static void parse_xml(INT32 args)
 {
@@ -2687,6 +2784,7 @@ static void parse_xml(INT32 args)
   struct pike_string *s;
   struct xmldata data;
   ONERROR e;
+  int doc_seq_pos = DOC_BEGINNING;
 
   if(args<2)
     Pike_error("Too few arguments to XML->parse()\n");
@@ -2713,6 +2811,7 @@ static void parse_xml(INT32 args)
   data.input.entity=0;
   data.input.callbackinfo=allocate_mapping(0);
   data.input.next=0;
+  CHECK_INPUT(data.input);
   data.func=sp+1-args;
   data.extra_args=sp+2-args;
   data.num_extra_args=args-2;
@@ -2724,7 +2823,11 @@ static void parse_xml(INT32 args)
   data.allow_pesmeg_everywhere=0;
 
   SET_ONERROR(e,free_xmldata, &data);
-  low_parse_xml(&data,0,1);
+  parse_optional_xmldecl(&data);
+  low_parse_xml(&data,0, &doc_seq_pos);
+  if (doc_seq_pos != DOC_AFTER_ROOT_ELEM &&
+      !(THIS->flags & COMPAT_ALLOW_7_6_ERRORS))
+    xmlerror ("Root element missing.", &data, NULL);
   CALL_AND_UNSET_ONERROR(e);
   tmp=*--sp;
   pop_n_elems(args);
@@ -2748,10 +2851,6 @@ static void free_xmldata(struct xmldata *data)
     
 }
 
-/*! @class XML
- *! @appears Parser.XML.Simple
- */
-
 /*! @decl void define_entity_raw(string entity, string raw)
  */
 static void define_entity_raw(INT32 args)
@@ -2788,6 +2887,7 @@ static void define_entity(INT32 args)
   data.input.to_free=0;
   data.input.entity=0;
   data.input.callbackinfo=allocate_mapping(0);
+  CHECK_INPUT(data.input);
   data.func=sp+2-args;
   data.extra_args=sp+3-args;
   data.num_extra_args=args-3;
@@ -2810,7 +2910,56 @@ static void define_entity(INT32 args)
 static void allow_rxml_entities(INT32 args)
 {
   check_all_args("XML->allow_rxml_entities", args, BIT_INT, 0);
-  THIS->rxml_mode = !!sp[-args].u.integer;
+  if (UNSAFE_IS_ZERO (sp - args))
+    THIS->flags &= ~ALLOW_RXML_ENTITIES;
+  else
+    THIS->flags |= ALLOW_RXML_ENTITIES;
+  pop_n_elems(args);
+  push_int(0);
+}
+
+/*! @decl void compat_allow_errors(string version)
+ *!
+ *! Set whether the parser should allow certain errors for
+ *! compatibility with earlier versions. @[version] can be:
+ *!
+ *! @string
+ *!   @value "7.2"
+ *!     Allow more data after the root element.
+ *!   @value "7.6"
+ *!     Allow multiple and invalidly placed "<?xml ... ?>" and
+ *!     "<!DOCTYPE ... >" declarations (invalid "<?xml ... ?>"
+ *!     declarations are otherwise treated as normal PI:s). Allow
+ *!     "<![CDATA[ ... ]]>" outside the root element. Allow the root
+ *!     element to be absent. This is the default in 7.6.
+ *! @endstring
+ *!
+ *! @[version] can also be zero to enable all error checks.
+ */
+static void compat_allow_errors(INT32 args)
+{
+  if (args < 1)
+    SIMPLE_TOO_FEW_ARGS_ERROR ("XML->compat_allow_errors", 1);
+
+  if (UNSAFE_IS_ZERO (sp - args))
+    THIS->flags &= ~(COMPAT_ALLOW_7_2_ERRORS|COMPAT_ALLOW_7_6_ERRORS);
+
+  else {
+    struct pike_string *str_7_2, *str_7_6;
+    MAKE_CONST_STRING (str_7_2, "7.2");
+    MAKE_CONST_STRING (str_7_6, "7.6");
+    if (sp[-args].type != T_STRING)
+      SIMPLE_BAD_ARG_ERROR ("XML->compat_allow_errors", 1, "string");
+    if (sp[-args].u.string == str_7_2)
+      THIS->flags |= COMPAT_ALLOW_7_2_ERRORS|COMPAT_ALLOW_7_6_ERRORS;
+    else if (sp[-args].u.string == str_7_6) {
+      THIS->flags &= ~COMPAT_ALLOW_7_2_ERRORS;
+      THIS->flags |= COMPAT_ALLOW_7_6_ERRORS;
+    }
+    else
+      Pike_error ("Got unknown version string.\n");
+  }
+
   pop_n_elems(args);
   push_int(0);
 }
@@ -2835,6 +2984,7 @@ static void parse_dtd(INT32 args)
   data.input.to_free=0;
   data.input.entity=0;
   data.input.callbackinfo=allocate_mapping(0);
+  CHECK_INPUT(data.input);
   data.func=sp+1-args;
   data.extra_args=sp+2-args;
   data.num_extra_args=args-2;
@@ -2878,6 +3028,7 @@ static void autoconvert(INT32 args)
 	pop_stack();
 	push_string(end_shared_string(t));
       }
+      /* FALL_THROUGH */
 	
       case 0xfeff: /* UTF-16, big-endian */
 	IF_XMLDEBUG(fprintf(stderr,"UTF-16, big endian detected.\n"));
@@ -3056,6 +3207,8 @@ static void init_xml_struct(struct object *o)
   THIS->is_cdata=sp[-1].u.mapping;
   sp--;
   dmalloc_touch_svalue(sp);
+
+  THIS->flags = COMPAT_ALLOW_7_6_ERRORS;
 }
 
 /*! @endclass
@@ -3079,8 +3232,6 @@ void init_xml(void)
 	       off + OFFSETOF(xmlobj, attributes),T_MAPPING);
   map_variable("__is_cdata", "mapping", ID_STATIC|ID_PRIVATE,
 	       off + OFFSETOF(xmlobj, is_cdata),T_MAPPING);
-  map_variable("__allow_rxml_entities", "int", ID_STATIC|ID_PRIVATE,
-	       off + OFFSETOF(xmlobj, rxml_mode), T_INT);
 
   set_init_callback(init_xml_struct);
 
@@ -3108,6 +3259,7 @@ void init_xml(void)
   ADD_FUNCTION("define_entity",define_entity,tFuncV(tStr tStr tMix,tMix,tVoid),0);
   ADD_FUNCTION("allow_rxml_entities", allow_rxml_entities,
 	       tFunc(tInt, tVoid), 0);
+  ADD_FUNCTION("compat_allow_errors", compat_allow_errors, tFunc(tStr, tVoid), 0);
   end_class("XML",0);
 
   ADD_FUNCTION("isbasechar",f_isBaseChar,tFunc(tInt,tInt),0);
