@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: acconfig.h,v 1.6 2003/05/03 17:48:22 mast Exp $
+|| $Id: acconfig.h,v 1.7 2005/03/21 15:29:26 grubba Exp $
 */
 
 /*
@@ -15,15 +15,10 @@
 #define __PIKE_SYBASE_CONFIG_H
 
 @TOP@
+
+/* Define to enable the sybase module. */
+#undef HAVE_SYBASE 
+
 @BOTTOM@
 
-/* End of automatic session. Doing stuff now */
-
-#undef PIKE_HAVE_LIBCOMN
-#undef PIKE_HAVE_LIBCS
-#undef PIKE_HAVE_LIBCT
-#undef PIKE_HAVE_LIBINTL
-#undef PIKE_HAVE_LIBSYBTCL
-#undef HAVE_SYBASE
-
 #endif /* __PIKE_SYBASE_CONFIG_H */
