@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: pike_memory.c,v 1.156 2005/06/22 16:00:54 grubba Exp $
+|| $Id: pike_memory.c,v 1.159 2006/07/05 01:10:26 mast Exp $
 */
 
 #include "global.h"
@@ -27,7 +27,7 @@
 
 #include <errno.h>
 
-RCSID("$Id: pike_memory.c,v 1.156 2005/06/22 16:00:54 grubba Exp $");
+RCSID("$Id: pike_memory.c,v 1.159 2006/07/05 01:10:26 mast Exp $");
 
 /* strdup() is used by several modules, so let's provide it */
 #ifndef HAVE_STRDUP
@@ -287,7 +287,8 @@ PMOD_EXPORT void *debug_xalloc(size_t size)
   ret=(void *)malloc(size);
   if(ret) return ret;
 
-  Pike_error("Out of memory.\n");
+  Pike_error("Out of memory - failed to allocate %"PRINTSIZET"d bytes.\n",
+	     size);
   return 0;
 }
 
@@ -346,7 +347,8 @@ char *debug_qalloc(size_t size)
   }
 #endif
 
-  Pike_fatal("Completely out of memory!\n");
+  Pike_fatal("Completely out of memory - "
+	     "failed to allocate %"PRINTSIZET"d bytes!\n", size);
   /* NOT_REACHED */
   return NULL;	/* Keep the compiler happy. */
 }
@@ -513,6 +515,11 @@ void mexec_free(void *ptr)
 	       ptr, mblk->magic, hdr);
   }
 #endif /* MEXEC_MAGIC */
+
+#ifdef VALGRIND_DISCARD_TRANSLATIONS
+  VALGRIND_DISCARD_TRANSLATIONS (mblk, mblk->size);
+#endif
+
   blk = (struct mexec_free_block *)mblk;
 
   next = hdr->free;
@@ -713,7 +720,43 @@ void *mexec_realloc(void *ptr, size_t sz)
 
   return res + 1;
 }
-#else
+
+#elif defined (VALGRIND_DISCARD_TRANSLATIONS)
+
+void *mexec_alloc (size_t sz)
+{
+  size_t *blk = malloc (sz + sizeof (size_t));
+  if (!blk) return NULL;
+  *blk = sz;
+  return blk + 1;
+}
+
+void *mexec_realloc (void *ptr, size_t sz)
+{
+  if (ptr) {
+    size_t *oldblk = ptr;
+    size_t oldsz = oldblk[-1];
+    size_t *newblk = realloc (oldblk - 1, sz + sizeof (size_t));
+    if (!newblk) return NULL;
+    if (newblk != oldblk)
+      VALGRIND_DISCARD_TRANSLATIONS (oldblk, oldsz);
+    else if (oldsz > sz)
+      VALGRIND_DISCARD_TRANSLATIONS (newblk + sz, oldsz - sz);
+    *newblk = sz;
+    return newblk + 1;
+  }
+  return mexec_malloc (sz);
+}
+
+void mexec_free (void *ptr)
+{
+  size_t *blk = ptr;
+  VALGRIND_DISCARD_TRANSLATIONS (blk, blk[-1]);
+  free (blk - 1);
+}
+
+#else  /* !(HAVE_MMAP && MEXEC_USES_MMAP) && !VALGRIND_DISCARD_TRANSLATIONS */
+
 void *mexec_alloc(size_t sz)
 {
   return malloc(sz);
@@ -727,7 +770,8 @@ void mexec_free(void *ptr)
 {
   free(ptr);
 }
-#endif /* HAVE_MMAP && MEXEC_USES_MMAP */
+
+#endif  /* !(HAVE_MMAP && MEXEC_USES_MMAP) && !VALGRIND_DISCARD_TRANSLATIONS */
 
 /* #define DMALLOC_TRACE */
 /* #define DMALLOC_TRACELOGSIZE	256*1024 */
