@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: b_source_normal_file.c,v 1.11 2004/04/04 01:02:01 nilsson Exp $
+|| $Id: b_source_normal_file.c,v 1.14 2005/05/18 15:13:35 mast Exp $
 */
 
 #include "global.h"
@@ -26,14 +26,6 @@
  * Argument: Stdio.File instance pointing to a normal file
  */
 
-#ifndef S_ISREG
-#ifdef S_IFREG
-#define S_ISREG(mode)	(((mode) & (S_IFMT)) == (S_IFREG))
-#else /* !S_IFREG */
-#define S_ISREG(mode)	(((mode) & (_S_IFMT)) == (_S_IFREG))
-#endif /* S_IFREG */
-#endif /* !S_ISREG */
-
 static struct program *Fd_ref_program = NULL;
 
 struct fd_source
@@ -42,10 +34,10 @@ struct fd_source
   struct object *obj;
   char buffer[CHUNK];
   int fd;
-  size_t len;
+  off_t len;
 };
 
-static struct data get_data( struct source *_s, ptrdiff_t len )
+static struct data get_data( struct source *_s, int len )
 {
   struct fd_source *s = (struct fd_source *)_s;
   struct data res;
@@ -55,10 +47,10 @@ static struct data get_data( struct source *_s, ptrdiff_t len )
   res.do_free = 0;
   res.off = 0;
   res.data = s->buffer;
-  
-  if( len > (ptrdiff_t)s->len )
+
+  if( len > s->len )
   {
-    len = (ptrdiff_t)s->len;
+    len = s->len;
     s->s.eof = 1;
   }
   THREADS_ALLOW();
@@ -69,7 +61,7 @@ static struct data get_data( struct source *_s, ptrdiff_t len )
 
   res.len = rr;
 
-  if( rr < len )
+  if( rr<0 || rr < len )
     s->s.eof = 1;
   return res;
 }
